import os
import subprocess
import sys
from pathlib import Path

import pytest

from snek import actions, api, cli, repo, shell, structure, toml
from snek.file_system import chdir, move, rm_rf


def test_init_commit_repo(tmpfolder):
    with tmpfolder.mkdir("my_porject").as_cwd():
        struct = {
            "my_file": "Some other content",
            "my_dir": {"my_file": "Some more content"},
            "dummy": None,
        }
        structure.create_structure(struct, {})
        dummy_file = Path("dummy")
        with dummy_file.open(mode="w"):
            os.utime(str(dummy_file), None)
        repo.init_commit_repo(".", struct)
        assert Path(".git").exists()


def test_pretend_init_commit_repo(tmpfolder):
    with tmpfolder.mkdir("my_porject").as_cwd():
        struct = {
            "my_file": "Some other content",
            "my_dir": {"my_file": "Some more content"},
            "dummy": None,
        }
        structure.create_structure(struct, {})
        dummy_file = Path("dummy")
        with dummy_file.open(mode="w"):
            os.utime(str(dummy_file), None)
        repo.init_commit_repo(".", struct, pretend=True)
        assert not Path(".git").exists()


def test_init_commit_repo_with_wrong_structure(tmpfolder):
    project = "my_project"
    struct = {"my_file": type("StrangeType", (object,), {})()}
    os.mkdir(project)
    with pytest.raises(TypeError):
        repo.init_commit_repo(project, struct)


def test_add_tag(tmpfolder):
    project = "my_project"
    struct = {
        "my_file": "Some other content",
        "my_dir": {"my_file": "Some more content"},
    }
    structure.create_structure(struct, dict(project_path=project))
    repo.init_commit_repo(project, struct)
    repo.add_tag(project, "v0.0")
    repo.add_tag(project, "v0.1", "Message with whitespace")


@pytest.mark.slow
def test_version_of_subdir(tmpfolder):
    projects = ["main_project", "inner_project"]
    for project in projects:
        opts = cli.parse_args([project])
        opts = api.bootstrap_options(opts)
        _, opts = actions.get_default_options({}, opts)
        struct, _ = structure.define_structure({}, opts)
        struct, _ = structure.create_structure(struct, opts)
        repo.init_commit_repo(project, struct)
    rm_rf(Path("inner_project", ".git"))
    move("inner_project", target="main_project/inner_project")

    # setuptools_scm required explicitly setting the git root when setup.py is
    # not at the root of the repository
    nested_setup_py = Path(tmpfolder, "main_project/inner_project/setup.py")
    content = nested_setup_py.read_text()
    content = content.replace(
        "use_scm_version={", 'use_scm_version={"root": "..", "relative_to": __file__, '
    )
    nested_setup_py.write_text(content)
    nested_pyproject_toml = Path(tmpfolder, "main_project/inner_project/pyproject.toml")
    config = toml.loads(nested_pyproject_toml.read_text())
    config["tool"]["setuptools_scm"]["root"] = ".."
    nested_pyproject_toml.write_text(toml.dumps(config))

    with chdir("main_project"):
        main_version = (
            subprocess.check_output([sys.executable, "setup.py", "--version"])
            .strip()
            .splitlines()[-1]
        )
        with chdir("inner_project"):
            inner_version = (
                subprocess.check_output([sys.executable, "setup.py", "--version"])
                .strip()
                .splitlines()[-1]
            )
    assert main_version.strip() == inner_version.strip()


def test_is_git_repo(tmpfolder):
    assert not repo.is_git_repo("/a-folder/that-not/exist")
    newdir = tmpfolder.join("new").ensure_dir()
    assert not repo.is_git_repo(str(newdir))
    newdir.chdir()
    shell.git("init")
    tmpfolder.chdir()
    assert repo.is_git_repo(str(newdir))


def test_get_git_root(tmpfolder):
    project = "my_project"
    struct = {
        "my_file": "Some other content",
        "my_dir": {"my_file": "Some more content"},
    }
    structure.create_structure(struct, {"project_path": project})
    repo.init_commit_repo(project, struct)
    with chdir(project):
        git_root = repo.get_git_root()
    assert Path(git_root).name == project


def test_get_git_root_with_nogit(tmpfolder, nogit_mock):
    project = "my_project"
    struct = {
        "my_file": "Some other content",
        "my_dir": {"my_file": "Some more content"},
    }
    structure.create_structure(struct, {"project_path": project})
    with chdir(project):
        git_root = repo.get_git_root(default=".")
    assert git_root == "."



from os.path import getmtime
from pathlib import Path
from tempfile import TemporaryDirectory
from textwrap import dedent

import pytest

from snek import cli, info, operations, structure, templates
from snek.actions import get_default_options
from snek.api import NO_CONFIG, bootstrap_options, create_project
from snek.exceptions import (
    DirectoryAlreadyExists,
    InvalidIdentifier,
    NoSnekProject,
)
from snek.extensions import Extension
from snek.file_system import chdir


def create_extension(*hooks):
    """Shorthand to define extensions from a list of actions"""

    class TestExtension(Extension):
        def activate(self, actions):
            for hook in hooks:
                actions = self.register(actions, hook, after="define_structure")
            return actions

    return TestExtension()


def test_create_project_call_extension_hooks(tmpfolder, git_mock):
    # Given an extension with hooks,
    called = []

    def pre_hook(struct, opts):
        called.append("pre_hook")
        return struct, opts

    def post_hook(struct, opts):
        called.append("post_hook")
        return struct, opts

    # when created project is called,
    create_project(
        project_path="proj", extensions=[create_extension(pre_hook, post_hook)]
    )

    # then the hooks should also be called.
    assert "pre_hook" in called
    assert "post_hook" in called


def test_create_project_generate_extension_files(tmpfolder, git_mock):
    # Given a blank state,
    assert not Path("proj/tests/extra.file").exists()
    assert not Path("proj/tests/another.file").exists()

    # and an extension with extra files,
    def add_files(struct, opts):
        struct = structure.ensure(struct, "tests/extra.file", "content")
        struct = structure.merge(struct, {"tests": {"another.file": "content"}})

        return struct, opts

    # when the created project is called,
    create_project(project_path="proj", extensions=[create_extension(add_files)])

    # then the files should be created
    assert Path("proj/tests/extra.file").exists()
    assert tmpfolder.join("proj/tests/extra.file").read() == "content"
    assert Path("proj/tests/another.file").exists()
    assert tmpfolder.join("proj/tests/another.file").read() == "content"


def test_create_project_respect_operations(tmpfolder, git_mock):
    # Given an existing project
    create_project(project_path="proj")
    for i in (0, 1, 3, 5, 6):
        tmpfolder.ensure("proj/tests/file" + str(i)).write("old")
        assert Path("proj/tests/file" + str(i)).exists()

    # and an extension with extra files
    def add_files(struct, opts):
        nov, sou = operations.no_overwrite(), operations.skip_on_update()
        struct = structure.ensure(struct, "tests/file0", "new")
        struct = structure.ensure(struct, "tests/file1", "new", nov)
        struct = structure.ensure(struct, "tests/file2", "new", sou)
        struct = structure.merge(
            struct,
            {
                "tests": {
                    "file3": ("new", nov),
                    "file4": ("new", sou),
                    "file5": ("new", operations.create),
                    "file6": "new",
                }
            },
        )

        return struct, opts

    # When the created project is called,
    create_project(
        project_path="proj", update=True, extensions=[create_extension(add_files)]
    )

    # then the NO_CREATE files should not be created,
    assert not Path("proj/tests/file2").exists()
    assert not Path("proj/tests/file4").exists()
    # the NO_OVERWRITE files should not be updated
    assert tmpfolder.join("proj/tests/file1").read() == "old"
    assert tmpfolder.join("proj/tests/file3").read() == "old"
    # and files with no rules or `None` rules should be updated
    assert tmpfolder.join("proj/tests/file0").read() == "new"
    assert tmpfolder.join("proj/tests/file5").read() == "new"
    assert tmpfolder.join("proj/tests/file6").read() == "new"


def test_create_project_when_folder_exists(tmpfolder, git_mock):
    tmpfolder.ensure("my-project", dir=True)
    opts = dict(project_path="my-project")
    with pytest.raises(DirectoryAlreadyExists):
        create_project(opts)
    opts = dict(project_path="my-project", force=True)
    create_project(opts)


def test_create_project_with_valid_package_name(tmpfolder, git_mock):
    opts = dict(project_path="my-project", package="my_package")
    create_project(opts)


def test_create_project_with_invalid_package_name(tmpfolder, git_mock):
    opts = dict(project_path="my-project", package="my:package")
    with pytest.raises(InvalidIdentifier):
        create_project(opts)


def test_create_project_when_updating(tmpfolder, git_mock):
    opts = dict(project_path="my-project")
    create_project(opts)
    opts = dict(project_path="my-project", update=True)
    create_project(opts)
    assert Path("my-project").exists()


def test_create_project_with_license(tmpfolder, git_mock):
    _, opts = get_default_options(
        {}, dict(project_path="my-project", license="BSD-3-Clause")
    )
    # ^ The entire default options are needed, since template
    #   uses computed information

    create_project(opts)
    assert Path("my-project").exists()
    content = tmpfolder.join("my-project/LICENSE.txt").read()
    assert content == templates.license(opts)


def test_api(tmpfolder):
    opts = dict(project_path="created_proj_with_api")
    create_project(opts)
    assert Path("created_proj_with_api").exists()
    assert Path("created_proj_with_api/.git").exists()


def test_pretend(tmpfolder):
    opts = dict(project_path="created_proj_with_api", pretend=True)
    create_project(opts)
    assert not Path("created_proj_with_api").exists()


def test_pretend_when_updating_does_not_make_changes(tmpfolder):
    # Given a project already exists
    opts = dict(project_path="proj", license="MIT")
    create_project(opts)

    setup_changed = getmtime("proj/setup.cfg")
    license_changed = getmtime("proj/LICENSE.txt")

    # When it is updated with different configuration,
    create_project(
        project_path="proj",
        update=True,
        force=True,
        pretend=True,
        url="my.project.net",
        license="MPL-2.0",
    )

    # Then nothing should change
    assert getmtime("proj/setup.cfg") == setup_changed
    assert "my.project.net" not in tmpfolder.join("proj/setup.cfg").read()

    assert getmtime("proj/LICENSE.txt") == license_changed
    assert "MIT License" in tmpfolder.join("proj/LICENSE.txt").read()


def test_bootstrap_opts_raises_when_updating_non_existing():
    with pytest.raises(NoSnekProject):
        bootstrap_options(project_path="non-existent", update=True)


def test_bootstrap_opts_raises_when_config_file_doesnt_exist():
    opts = dict(project_path="non-existent", config_files=["non-existent.cfg"])
    with pytest.raises(FileNotFoundError):
        bootstrap_options(opts)


def test_bootstrap_using_config_file(tmpfolder):
    # First we create a config file
    opts = dict(project_path="proj", name="my-proj", license="MPL-2.0")
    opts = bootstrap_options(opts)
    _, opts = get_default_options({}, opts)
    setup_cfg = Path(str(tmpfolder.join("setup.cfg")))
    setup_cfg.write_text(templates.setup_cfg(opts))

    # Then we input this configfile to the API
    new_opts = dict(project_path="another", config_files=[str(setup_cfg)])
    new_opts = bootstrap_options(new_opts)

    # Finally, the bootstraped options should contain the same values
    # as the given config file
    assert new_opts["name"] == "my-proj"
    assert new_opts["package"] == "my_proj"
    assert new_opts["license"] == "MPL-2.0"
    assert str(new_opts["project_path"]) == "another"
    assert all(k in new_opts for k in "author email url".split())


DEFAULT_CONFIG = """\
[metadata]
author = John Doe
author-email = john.joe@gmail.com

[snek]
extensions =
    namespace
    cirrus
namespace = my_namespace.my_sub_namespace
"""


@pytest.fixture
def with_default_config(fake_config_dir):
    cfg = fake_config_dir / info.CONFIG_FILE
    cfg.write_text(DEFAULT_CONFIG)

    yield cfg


def test_bootstrap_with_default_config(tmpfolder, with_default_config):
    # Given a default config file exists and contains stuff
    _ = with_default_config
    # when bootstrapping options
    opts = dict(project_path="xoxo")
    new_opts = bootstrap_options(opts)
    # the stuff will be considered
    assert new_opts["author"] == "John Doe"
    assert new_opts["email"] == "john.joe@gmail.com"
    assert new_opts["namespace"] == "my_namespace.my_sub_namespace"
    extensions = new_opts["extensions"]
    assert len(extensions) == 2
    extensions_names = sorted(e.name for e in extensions)
    assert " ".join(extensions_names) == "cirrus namespace"


def test_bootstrap_with_no_config(tmpfolder, with_default_config):
    # Given a default config file exists and contains stuff
    _ = with_default_config
    # when bootstrapping options with NO_CONFIG
    opts = dict(project_path="xoxo", config_files=NO_CONFIG)
    new_opts = bootstrap_options(opts)
    # the stuff will not be considered
    assert new_opts.get("author") != "John Doe"
    assert new_opts.get("email") != "john.joe@gmail.com"
    assert new_opts.get("namespace") != "my_namespace.my_sub_namespace"
    extensions = new_opts.get("extensions", [])
    assert len(extensions) != 2
    extensions_names = sorted(e.name for e in extensions)
    assert " ".join(extensions_names) != "cirrus namespace"


@pytest.fixture
def macos_config_dir(tmp_path, monkeypatch, fake_config_dir):
    with TemporaryDirectory(prefix="config", dir=str(tmp_path)) as tmp:
        confdir = Path(tmp, "Library/Application Support/snek")

    _ = fake_config_dir  # included for reproducible ordering of fixture evaluation
    monkeypatch.setattr("snek.info.config_dir", lambda *_, **__: confdir)
    yield confdir


def test_bootstrap_with_legacy_macos_default_config(
    monkeypatch, tmpfolder, macos_config_dir
):
    monkeypatch.setattr("sys.platform", "darwin_test")  # pretend to be on macOS
    # Given a default config file in an outdated config dir on macOS
    # (see platformdirs v3.0.0)
    new_dir = macos_config_dir
    assert not new_dir.exists()  # sanity check
    old_dir = info._old_macos_config_dir(new_dir)
    old_dir.mkdir(parents=True, exist_ok=False)
    (old_dir / "default.cfg").write_text(DEFAULT_CONFIG)
    # when bootstrapping options
    opts = dict(project_path="xoxo")
    new_opts = bootstrap_options(opts)
    # the old directory will be moved to the new location
    assert new_dir.exists()
    assert not old_dir.exists()
    # and the stuff in the old config will be considered
    assert new_opts["author"] == "John Doe"
    assert new_opts["email"] == "john.joe@gmail.com"
    assert new_opts["namespace"] == "my_namespace.my_sub_namespace"
    extensions = new_opts["extensions"]
    assert len(extensions) == 2
    extensions_names = sorted(e.name for e in extensions)
    assert " ".join(extensions_names) == "cirrus namespace"


def test_create_project_with_default_config(tmpfolder, with_default_config):
    # Given a default config file exists and contains stuff
    _ = with_default_config
    project = Path(str(tmpfolder)) / "xoxo"
    # when a new project is created
    create_project(project_path="xoxo", name="project")
    # then the default config is considered
    assert (project / "src/my_namespace/my_sub_namespace/project").exists()
    assert (project / "tox.ini").exists()
    assert (project / ".cirrus.yml").exists()


@pytest.fixture
def with_existing_proj_config(tmp_path):
    proj = tmp_path / "proj"
    proj.mkdir(parents=True, exist_ok=True)
    (proj / "setup.cfg").write_text(
        dedent(
            """\
            [metadata]
            name = SuperProj
            description = some text
            author = John Doe
            author-email = john.doe@example.com
            url = www.example.com
            license = gpl3

            [snek]
            package = super_proj
            """
        )
    )
    with chdir(str(proj)):
        yield proj


def test_options_with_existing_proj_config_and_cli(with_existing_proj_config):
    # Given an existing project with a setup.cfg
    _ = with_existing_proj_config
    # When the CLI is called with no extra parameters
    opts = cli.parse_args(["--update", "."])
    opts = bootstrap_options(opts)
    _, opts = get_default_options({}, opts)

    # After all the opt processing actions are finished
    # The parameters in the old setup.py files are preserved
    assert opts["name"] == "SuperProj"
    assert opts["description"] == "some text"
    assert opts["author"] == "John Doe"
    assert opts["email"] == "john.doe@example.com"
    assert opts["url"] == "www.example.com"
    assert opts["license"] == "GPL-3.0-only"
    assert opts["package"] == "super_proj"



"""Thin wrapper around the dependency so we can maintain some stability while being able
to swap implementations (e.g. replace `tomlkit`_ with `toml`_).

`tomlkit`_ is preferred because it supports comments.

.. _tomlkit: https://github.com/sdispater/tomlkit
.. _toml: https://github.com/uiri/toml
"""
from typing import Any, List, MutableMapping, NewType, Tuple, TypeVar, Union, cast

import tomlkit

TOMLMapping = NewType("TOMLMapping", MutableMapping)
"""Abstraction on the value returned by :obj:`loads`.

This kind of object ideally should present a dict-like interface and be able to preserve
the formatting and comments of the original TOML file.
"""

T = TypeVar("T")


def loads(text: str) -> TOMLMapping:
    """Parse a string containing TOML into a dict-like object,
    preserving style somehow.
    """
    return TOMLMapping(cast(MutableMapping, tomlkit.loads(text)))


def dumps(obj: Union[TOMLMapping, dict]) -> str:
    # and do not change the object
    # TODO: If/when tomlkit allows serialising arbitrary Mapping objects, replace union
    #       with Mapping
    """Serialize a dict-like object into a TOML str,
    If the object was generated via :obj:`loads`, then the style will be preserved.
    """
    return tomlkit.dumps(obj)  # type: ignore[arg-type]
    # TODO: Once tomlkit improves dumps' type hints, remove type ignore comment


def setdefault(obj: MutableMapping, key: str, value: T) -> T:
    # TODO: Use recursive type for obj, when available in mypy, and remove type: ignore
    """tomlkit seems to be tricky to use together with setdefault, this function is a
    workaround for that.

    When ``key`` is string containing ``'.'``, it will perform a nested setdefault.
    """
    keys = key.split(".")
    items: List[Tuple[str, Any]] = list(zip(keys, [{}] * len(keys)))
    items[-1] = (keys[-1], value)

    last = obj
    for key, value in items:
        if key not in last:
            last[key] = value
        last = last[key]  # type: ignore

    return cast(T, last)



"""
Custom logging infrastructure to provide execution information for the user.
"""
import logging
from collections import defaultdict
from contextlib import contextmanager
from logging import INFO, Formatter, Handler, LoggerAdapter, StreamHandler, getLogger
from os.path import realpath, relpath
from os.path import sep as pathsep
from typing import DefaultDict, Optional, Sequence, cast

from . import termui

DEFAULT_LOGGER = __name__
"""Name of snek's default logger (it can be used with :obj:`logging.getLogger`)"""

Styles = Sequence[str]


def _are_equal_paths(path1, path2):
    return realpath(str(path1)) == realpath(str(path2))


def _is_current_path(path):
    return _are_equal_paths(path, ".")


class ReportFormatter(Formatter):
    """Formatter that understands custom fields in the log record."""

    ACTIVITY_MAXLEN = 12
    SPACING = "  "
    CONTEXT_PREFIX = "from"
    TARGET_PREFIX = "to"

    def format(self, record):
        """Compose message when a record with report information is given."""
        if hasattr(record, "activity"):
            return self.format_report(record)

        return self.format_default(record)

    def create_padding(self, activity):
        """Create the appropriate padding in order to align activities."""
        actual = len(activity)
        count = max(self.ACTIVITY_MAXLEN - actual, 0)
        return " " * count

    def format_path(self, path):
        """Simplify paths to avoid wasting space in terminal."""
        from .file_system import is_pathname_valid  # late import due to cycles

        # TODO: Rather handle Path objects instead converting to str
        path = str(path)

        if is_pathname_valid(path) and pathsep in path:
            # Heuristic to determine if subject is a file path
            # that needs to be made short
            abbrev = relpath(path)

            if len(abbrev) < len(path):
                # Ignore if not shorter
                path = abbrev

        return path

    def format_activity(self, activity):
        """Format the activity keyword."""
        return activity

    # Subclasses may need the activity name to choose correct format,
    # so the following 3 methods accept a second parameter
    # (even if they not use it)
    def format_subject(self, subject, _activity=None):
        """Format the subject of the activity."""
        return self.format_path(subject) if subject else ""

    def format_target(self, target, _activity=None):
        """Format extra information about the activity target."""
        if target and not _is_current_path(target):
            return f"{self.TARGET_PREFIX} '{self.format_path(target)}'"

        return ""

    def format_context(self, context, _activity=None):
        """Format extra information about the activity context."""
        if context and not _is_current_path(context):
            return f"{self.CONTEXT_PREFIX} '{self.format_path(context)}'"

        return ""

    def format_default(self, record):
        """Format default log messages."""
        record.msg = self.SPACING * max(record.nesting, 0) + record.msg

        return super().format(record)

    def format_report(self, record):
        """Compose message when a custom record is given."""
        activity = record.activity
        record.msg = (
            self.create_padding(activity)
            + self.format_activity(activity)
            + self.SPACING * max(record.nesting + 1, 0)
            + " ".join(
                [
                    text
                    for text in [
                        self.format_subject(record.subject, activity),
                        self.format_target(record.target, activity),
                        self.format_context(record.context, activity),
                    ]
                    if text  # Filter empty strings
                ]
            )
        )

        return super().format(record)


class ColoredReportFormatter(ReportFormatter):
    """Format logs with ANSI colors."""

    ACTIVITY_STYLES: DefaultDict[str, Styles] = defaultdict(
        lambda: ("blue", "bold"),
        create=("green", "bold"),
        move=("green", "bold"),
        remove=("red", "bold"),
        delete=("red", "bold"),
        skip=("yellow", "bold"),
        run=("magenta", "bold"),
        invoke=("bold",),
    )

    SUBJECT_STYLES: DefaultDict[str, Styles] = defaultdict(tuple, invoke=("blue",))

    LOG_STYLES: DefaultDict[str, Styles] = defaultdict(
        tuple,
        debug=("green",),
        info=("blue",),
        warning=("yellow",),
        error=("red",),
        critical=("red", "bold"),
    )

    CONTEXT_PREFIX = termui.decorate(ReportFormatter.CONTEXT_PREFIX, "magenta", "bold")

    TARGET_PREFIX = termui.decorate(ReportFormatter.TARGET_PREFIX, "magenta", "bold")

    def format_activity(self, activity):
        return termui.decorate(activity, *self.ACTIVITY_STYLES[activity])

    def format_subject(self, subject, activity=None):
        subject = super().format_subject(subject, activity)
        return termui.decorate(subject, *self.SUBJECT_STYLES[activity])

    def format_default(self, record):
        record.msg = termui.decorate(
            record.msg, *self.LOG_STYLES[record.levelname.lower()]
        )
        return super().format_default(record)


class ReportLogger(LoggerAdapter):
    """Suitable wrapper for snek CLI interactive execution reports.

    Args:
        logger (logging.Logger): custom logger to be used. Optional: the
            default logger will be used.
        handlers (logging.Handler): custom logging handler to be used.
            Optional: a :class:`logging.StreamHandler` is used by default.
        formatter (logging.Formatter): custom formatter to be used.
            Optional: by default a :class:`~.ReportFormatter` is created and
            used.
        extra (dict): extra attributes to be merged into the log record.
            Options, empty by default.
        propagate (bool): whether or not to propagate messages in the logging hierarchy,
            ``False`` by default. See :obj:`logging.Logger.propagate`.

    Attributes:
        nesting (int): current nesting level of the report.
    """

    def __init__(
        self,
        logger: Optional[logging.Logger] = None,
        handler: Optional[logging.Handler] = None,
        formatter: Optional[logging.Formatter] = None,
        extra: Optional[dict] = None,
        propagate=False,
    ):
        self.nesting = 0
        self._wrapped: logging.Logger = logger or getLogger(DEFAULT_LOGGER)
        self.propagate = propagate
        self.extra = extra or {}
        self.handler = handler or StreamHandler()
        self.formatter = formatter or ReportFormatter()
        super().__init__(self._wrapped, self.extra)

    @property
    def propagate(self) -> bool:
        """Whether or not to propagate messages in the logging hierarchy,
        See :obj:`logging.Logger.propagate`.
        """
        return self._propagate

    @propagate.setter
    def propagate(self, value: bool):
        self._propagate = value
        self._wrapped.propagate = value

    @property
    def wrapped(self) -> logging.Logger:
        """Underlying logger object"""
        return self._wrapped

    @wrapped.setter
    def wrapped(self, value: logging.Logger):
        self._wrapped = value
        value.propagate = self.propagate
        self.handler = cast(Handler, getattr(self, "_handler", None))

    @property
    def handler(self) -> logging.Handler:
        """Stream handler configured for providing user feedback in snek CLI"""
        return self._handler

    @handler.setter
    def handler(self, value: logging.Handler):
        self._handler = value
        self._wrapped.handlers.clear()
        self._wrapped.addHandler(self._handler)

    @property
    def formatter(self) -> logging.Formatter:
        """Formatter configured in the default handler"""
        return self._formatter

    @formatter.setter
    def formatter(self, value: logging.Formatter):
        self._formatter = value
        self.handler.setFormatter(value)

    @property
    def level(self):
        """Effective level of the logger"""
        return self.wrapped.getEffectiveLevel()

    @level.setter
    def level(self, value):
        """Set the logger level"""
        self.wrapped.setLevel(value)

    def process(self, msg, kwargs):
        """Method overridden to augment LogRecord with the `nesting` attribute"""
        (msg, kwargs) = super().process(msg, kwargs)
        extra = kwargs.get("extra", {})
        extra["nesting"] = self.nesting
        kwargs["extra"] = extra
        return msg, kwargs

    def report(
        self, activity, subject, context=None, target=None, nesting=None, level=INFO
    ):
        """Log that an activity has occurred during scaffold.

        Args:
            activity (str): usually a verb or command, e.g. ``create``,
                ``invoke``, ``run``, ``chdir``...
            subject (str or os.PathLike): usually a path in the file system or
                an action identifier.
            context (str or os.PathLike): path where the activity take place.
            target (str or os.PathLike): path affected by the activity
            nesting (int): optional nesting level. By default it is calculated
                from the activity name.
            level (int): log level. Defaults to :obj:`logging.INFO`.
                See :mod:`logging` for more information.

        Notes:
            This method creates a custom log record, with additional fields:
            **activity**, **subject**, **context**, **target** and **nesting**,
            but an empty **msg** field. The :class:`ReportFormatter`
            creates the log message from the other fields.

            Often **target** and **context** complement the logs when
            **subject** does not hold all the necessary information. For
            example::

                logger.report('copy', 'my/file', target='my/awesome/path')
                logger.report('run', 'command', context='current/working/dir')
        """
        return self.wrapped.log(
            level,
            "",
            extra={
                "activity": activity,
                "subject": subject,
                "context": context,
                "target": target,
                "nesting": nesting or self.nesting,
            },
        )

    @contextmanager
    def indent(self, count=1):
        """Temporarily adjust padding while executing a context.

        Example:

            .. code-block:: python

                from snek.log import logger

                logger.report("invoke", "custom_action")
                with logger.indent():
                    logger.report("create", "some/file/path")

                # Expected logs:
                # --------------------------------------
                #       invoke  custom_action
                #       create    some/file/path
                # --------------------------------------
                # Note how the spacing between activity and subject in the
                # second entry is greater than the equivalent in the first one.

        Note:
            This method is not thread-safe and should be used with care.
        """
        prev = self.nesting
        self.nesting += count
        try:
            yield
        finally:
            self.nesting = prev

    def copy(self):
        """Produce a copy of the wrapped logger.

        Sometimes, it is better to make a copy of th report logger to keep
        indentation consistent.
        """
        clone = self.__class__(
            self.wrapped, self.handler, self.formatter, self.extra, self.propagate
        )
        clone.nesting = self.nesting

        return clone

    def reconfigure(self, opts: Optional[dict] = None, **kwargs):
        """Reconfigure some aspects of the logger object.

        Args:
            opts (dict): dict with the same elements as the keyword arguments below

        Keyword Args:
            log_level: One of the log levels specified in the :obj:`logging` module.
            use_colors: automatically set a colored formatter to the logger
                if ANSI codes support is detected. (Defaults to `True`).

        Additional keyword arguments will be ignored.
        """
        opts = (opts or {}).copy()
        opts.update(kwargs)

        if "log_level" in opts:
            self.level = opts["log_level"]

        # if terminal supports, use colors
        stream = getattr(self.handler, "stream", None)
        if opts.get("use_colors", True) and termui.supports_color(stream):
            self.formatter = ColoredReportFormatter()
            self.handler.setFormatter(self.formatter)

        return self


logger = ReportLogger()
"""Default logger configured for snek."""



import argparse
from textwrap import dedent
from unittest.mock import Mock

from snek import api, cli
from snek.extensions import config, interactive

from ..helpers import ArgumentParser


def normalise(text: str) -> str:
    return "\n".join(text.strip().splitlines())


def test_wrap():
    text = "as eireir tueroue asdiuodsi usaoifdu asdouusa doudas"
    wrapped = normalise(interactive.wrap(text, 20))
    assert wrapped == "as eireir tueroue\nasdiuodsi usaoifdu\nasdouusa doudas"


def test_comment():
    assert interactive.comment("a") == "# a"
    assert interactive.comment("a\nb") == "# a\n# b"
    assert interactive.comment("a\nb", indent_level=4) == "    # a\n    # b"
    assert interactive.comment("a\nb", comment_mark=";") == "; a\n; b"


def test_join_block():
    assert interactive.join_block("a", "", "") == "a"
    assert interactive.join_block("", "", "c") == "c"
    assert normalise(interactive.join_block("a", "", "c")) == "a\nc"
    assert normalise(interactive.join_block("", "b", "c")) == "b\nc"
    assert normalise(interactive.join_block("a", "b", "c")) == "a\nb\nc"


def test_long_option():
    for flags in (["-o", "--option"], ["--option", "-o"]):
        action = argparse.Action(flags, "dest")
        assert interactive.long_option(action).strip() == "--option"


def test_alternative_flags():
    for flags in (["-o", "--option", "-b"], ["--option", "-b", "-o"]):
        action = argparse.Action(flags, "dest")
        text = interactive.alternative_flags(action)
        assert "--option" not in text
        assert all([flag in text for flag in ("-o", "-b")])


def test_example_no_value():
    parser = ArgumentParser(conflict_handler="resolve")

    # When store_true option value is True, then it should not be commented
    action = parser.add_argument("--option", action="store_true")
    option_line = interactive.example_no_value(parser, action, {"option": True})
    assert option_line.strip() == "--option"
    option_line = interactive.example_no_value(parser, action, {"option": False})
    assert option_line.strip() == "# --option"

    # When store_false option value is False, then it should not be commented
    action = parser.add_argument("--option", action="store_false")
    option_line = interactive.example_no_value(parser, action, {"option": False})
    assert option_line.strip() == "--option"
    option_line = interactive.example_no_value(parser, action, {"option": True})
    assert option_line.strip() == "# --option"

    # When store_const option value is const, then it should not be commented
    action = parser.add_argument("--option", action="store_const", const=44, default=33)
    option_line = interactive.example_no_value(parser, action, {"option": 44})
    assert option_line.strip() == "--option"
    option_line = interactive.example_no_value(parser, action, {"option": 33})
    assert option_line.strip() == "# --option"

    # When no value is available in opts, then it should be commented
    action = argparse.Action(["--option", "-o"], "option", nargs=0)
    option_line = interactive.example_no_value(parser, action, {})
    assert option_line.strip() == "# --option"

    # When an extension is available, then it should not be commented
    option_line = interactive.example_no_value(
        parser, action, {"extensions": [Mock(flag="--option")]}
    )
    assert option_line.strip() == "--option"


def test_example_with_help():
    parser = ArgumentParser()
    action = parser.add_argument(
        "-o", "--option", action="store_true", help="do 42 things"
    )
    parser = ArgumentParser()
    text = """\
    # --option
        # (or alternatively: -o)
        # do 42 things
    """
    text = dedent(text)
    example = interactive.example_with_help(parser, action, {})
    assert normalise(dedent(text)) == normalise(example)

    # When option value is True, then it should not be commented
    text = """\
    --option
        # (or alternatively: -o)
        # do 42 things
    """
    example = interactive.example_with_help(parser, action, {"option": True})
    assert normalise(dedent(text)) == normalise(example)


def make_action(
    flags=("--option", "-o"),
    dest="option",
    nargs=None,
    help="do 42 things",
    metavar="OPTION",
):
    return argparse.Action(flags, dest, nargs=nargs, help=help, metavar=metavar)


def test_example():
    parser = ArgumentParser()

    # Options with variable nargs
    action = make_action(nargs=1)
    assert interactive.example(parser, action, {}).strip() == "# --option OPTION"
    assert interactive.example(parser, action, {"option": 32}).strip() == "--option 32"

    action = make_action(nargs=3)
    example = interactive.example(parser, action, {}).strip()
    assert example == "# --option OPTION OPTION OPTION"
    example = interactive.example(parser, action, {"option": [32, 21, 5]}).strip()
    assert example == "--option 32 21 5"

    action = make_action(nargs="*")
    example = interactive.example(parser, action, {}).strip()
    expected = ("# --option [OPTION [OPTION ...]]", "# --option [OPTION ...]")
    assert example in expected

    action = make_action(nargs="+")
    example = interactive.example(parser, action, {}).strip()
    assert example == "# --option OPTION [OPTION ...]"

    action = make_action(nargs="?")
    assert interactive.example(parser, action, {}).strip() == "# --option [OPTION]"

    # Positional argument:
    action = argparse.Action([], "arg", nargs=1, metavar="ARGUMENT")
    assert interactive.example(parser, action, {}).strip() == "# ARGUMENT"
    assert interactive.example(parser, action, {"arg": "value"}).strip() == "value"


def test_all_examples():
    parser = ArgumentParser()
    actions = [
        make_action(nargs=1),
        make_action(["--abc"], "abc", metavar="ABC", help="Abc-foobarize your project"),
    ]
    text = """\
    --option 23
        # (or alternatively: -o)
        # do 42 things


    # --abc ABC
        # Abc-foobarize your project
    """
    example = interactive.all_examples(parser, actions, {"option": 23})
    assert normalise(dedent(text)) == normalise(example)


def test_no_empty_example():
    # As reported on #333, the config flag was generating an empty example when the user
    # does not have a "{$XDG_CONFIG_HOME:-$HOME/.config}/snek/default.cfg" file,
    # which in turn generated the following error:
    #
    #    putup: error: argument --config: expected at least one argument
    #
    # This is a regression test to prevent that.

    # Explicit test case for the config example
    parser = ArgumentParser()
    extension = config.Config()
    extension.augment_cli(parser)
    action = next(a for a in parser._actions if "--config" in a.option_strings)
    example = interactive.example(parser, action, {"config_files": []})
    assert "# --config CONFIG_FILE" in normalise(example)  # example should be commented

    # Generalised test case
    parser = ArgumentParser()
    action = parser.add_argument("-x", dest="x", metavar="Y", nargs="+")
    example = interactive.example(parser, action, {"x": []})
    assert "# -x Y" in normalise(example)  # example should be commented
    example = interactive.example(parser, action, {"x": None})
    assert "# -x Y" in normalise(example)  # example should be commented


def test_multiple_options_same_dest():
    # General examples
    parser = ArgumentParser()
    x = parser.add_argument("-x", dest="val", action="store_const", const="x")
    y = parser.add_argument("-y", dest="val", action="store_const", const="y")

    # example should be commented when the value does not match
    example = normalise(interactive.example(parser, y, {"val": "x"}))
    assert "# -y" in example

    # example should be not commented when the value does match
    example = normalise(interactive.example(parser, y, {"val": "y"}))
    assert example.startswith("-y")

    # example should be commented when the value does not match
    example = normalise(interactive.example(parser, x, {"val": "y"}))
    assert "# -x" in example

    # example should be not commented when the value does match
    example = normalise(interactive.example(parser, x, {"val": "x"}))
    assert example.startswith("-x")

    # AD HOC examples found during debug
    # Just one of the 2 (`--config` or `--no-config`) should be activated in the example
    parser = ArgumentParser()
    extension = config.Config()
    extension.augment_cli(parser)

    nocfg = next(a for a in parser._actions if "--no-config" in a.option_strings)
    cfg = next(a for a in parser._actions if "--config" in a.option_strings)

    example = interactive.example(parser, nocfg, {"config_files": api.NO_CONFIG})
    assert normalise(example).startswith("--no-config")

    example = interactive.example(parser, cfg, {"config_files": api.NO_CONFIG})
    assert "# --config" in normalise(example)

    example = interactive.example(parser, cfg, {"config_files": ["file"]})
    assert normalise(example).startswith("--config file")

    example = interactive.example(parser, nocfg, {"config_files": ["file"]})
    assert "# --no-config" in normalise(example)


def test_commented_extension(monkeypatch):
    # When a flag is marked as commented
    config = {"comment": ["--option"], "ignore": []}
    monkeypatch.setattr(interactive, "get_config", lambda x: config[x])

    # And an extension corresponds to that flag
    parser = ArgumentParser()
    fake_extension = Mock(flag="--option")
    action = parser.add_argument(
        "--option", dest="extensions", action="append_const", const=fake_extension
    )
    option_line = interactive.example_no_value(
        parser, action, {"extensions": [fake_extension]}
    )
    # then it should be commented in the file
    assert option_line.strip() == "# --option"


def test_ignored_extension(monkeypatch):
    # When a flag is marked as ignored
    config = {"ignore": ["--option"], "comment": []}
    monkeypatch.setattr(interactive, "get_config", lambda x: config[x])

    # And an extension corresponds to that flag
    parser = ArgumentParser()
    fake_extension = Mock(flag="--option")
    action = parser.add_argument(
        "--option", dest="extensions", action="append_const", const=fake_extension
    )
    text = interactive.all_examples(parser, [action], {"extensions": [fake_extension]})
    # then it should be omitted
    assert "--option" not in text


def test_get_config():
    ignore = interactive.get_config("ignore")
    assert "--help" in ignore
    assert "--version" in ignore
    assert "--interactive" in ignore
    comment = interactive.get_config("comment")
    assert "--verbose" in comment
    assert "--very-verbose" in comment


def test_putup_real_examples():
    parser = ArgumentParser()
    cli.add_default_args(parser)
    for extension in cli.list_all_extensions():
        extension.augment_cli(parser)

    actions = interactive.get_actions(parser)
    text = normalise(interactive.all_examples(parser, actions, {}))
    assert "# --force" in text
    assert "# --update" in text
    assert "# --namespace" in text
    assert "# --no-tox" in text
    assert "--interactive" not in text
    assert "--help" not in text
    assert "--version" not in text


def test_cli(monkeypatch, tmpfolder):
    # When the user edit the contents of the file
    fake_content = """\
    myproj_path
    --name myproj
    --license gpl3
    --no-config
    # --namespace myns
    # ^  test commented options
    """
    fake_edit = tmpfolder / "snek.args"
    fake_edit.write_text(dedent(fake_content), "utf-8")
    monkeypatch.setattr("snek.shell.edit", lambda *_, **__: fake_edit)

    # Then, the options in the file should take place, not the ones given in the cli
    args = [
        "-vv",
        "--no-config",
        "--interactive",
        "myproj",
        "--no-tox",
        "--license",
        "mpl",
    ]
    cli.run(args)
    assert not (tmpfolder / "myproj").exists()
    assert (tmpfolder / "myproj_path/tox.ini").exists()
    assert (tmpfolder / "myproj_path/src/myproj/__init__.py").exists()
    license = (tmpfolder / "myproj_path/LICENSE.txt").read_text("utf-8")
    assert "GNU GENERAL PUBLIC LICENSE" in license
    assert "Version 3" in license
    # Commented options (or options that were not mentioned) should not take place
    assert not (tmpfolder / "myproj_path/src/myns").exists()
    assert not (tmpfolder / "myproj_path/.pre-commit-config.yaml").exists()



from pathlib import Path
from shutil import copytree

import pytest

from snek import cli
from snek import file_system as fs
from snek import shell
from snek.exceptions import ExtensionNotFound


def copy_example(name, target_dir) -> Path:
    example = Path(__file__).parent / "examples" / name
    target = Path(target_dir, name)
    return Path(copytree(str(example), str(target)))


def test_issue506(tmpfolder):
    try:
        # Given a project exists with the same `setup.cfg` from issue506
        with fs.chdir(copy_example("issue-506", tmpfolder)):
            shell.git("init", ".")
            shell.git("add", ".")
            shell.git("commit", "-m", "Add code before update")
            # when we run the update
            cli.main([".", "--update", "--force"])
            # then no error should be raised when updating
            # and the extension should load correctly (e.g. namespace)
            assert Path("src/snekext/company/__init__.py").exists()
            assert not Path("src/snekext/__init__.py").exists()
            assert not Path("src/company").exists()
    except ExtensionNotFound as ex:
        pytest.skip(f"Test cannot succeed without extensions.\n{ex}")



"""CLI options for using/saving preferences as snek config files."""
import argparse
from pathlib import Path
from typing import TYPE_CHECKING, List

from configupdater import ConfigUpdater

from .. import api, info, operations, templates
from . import Extension, store_with

if TYPE_CHECKING:  # pragma: no cover
    from ..actions import Action, ActionParams, ScaffoldOpts, Structure


class Config(Extension):
    """Add a few useful options for creating/using snek config files."""

    persist = False

    def augment_cli(self, parser: argparse.ArgumentParser):
        default_file = info.config_file(default=None)
        default_help = f" (defaults to: {default_file})" if default_file else ""
        group = parser.add_mutually_exclusive_group()
        group.add_argument(
            "--config",
            dest="config_files",
            metavar="CONFIG_FILE",
            nargs="+",
            type=Path,
            help=f"config file to read snek's preferences{default_help}",
        )
        group.add_argument(
            "--no-config",
            dest="config_files",
            action="store_const",
            const=api.NO_CONFIG,
            help="prevent snek from reading its default config file",
        )
        parser.add_argument(
            "--save-config",
            dest="save_config",
            action=store_with(self),
            nargs="?",
            const=default_file,
            default=argparse.SUPPRESS,
            type=Path,
            help=f"save the given options in a config file{default_help}",
        )
        return self

    def activate(self, actions: List["Action"]) -> List["Action"]:
        return actions[:-1] + [save, actions[-1]]  # Just before the last action


def save(struct: "Structure", opts: "ScaffoldOpts") -> "ActionParams":
    """Save the given opts as preferences in a snek's config file."""
    config = ConfigUpdater()

    if not opts.get("save_config"):
        file = info.config_file()
    else:
        file = Path(opts["save_config"])

    if file.exists():
        config.read(str(file), encoding="utf-8")
    else:
        config.read_string(
            "# snek's configuration file, see:\n"
            "# https://snek.org/en/latest/configuration.html\n#\n"
            "# Accepted in `metadata`: author, author_email and license.\n"
            "# Accepted in `snek`: extensions (and associated opts).\n"
        )

    if "metadata" not in config:
        config.add_section("metadata")

    # We will add metadata just if they are not the default ones
    metadata = config["metadata"]
    defaults = [
        ("author", opts["author"], info.username()),
        ("author_email", opts["email"], info.email()),
        ("license", opts["license"], api.DEFAULT_OPTIONS["license"]),
    ]

    metadata.update({k: v for k, v, default in defaults if v != default})
    templates.add_snek(config, opts)

    operations.create(file, str(config), opts)  # operations provide logging and pretend

    return struct, opts



from importlib.metadata import PackageNotFoundError, version  # pragma: no cover

try:
    __version__ = version(__name__)
except PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"



#!/usr/bin/env python
import logging
import sys
from pathlib import Path
from unittest.mock import Mock

from snek import shell
from snek.api import create_project
from snek.cli import run
from snek.extensions import pre_commit
from snek.templates import get_template


def assert_in_logs(caplog, *expected):
    log = caplog.text
    for text in expected:
        assert text in log


# ---- "Isolated" tests ----


def test_find_executable(monkeypatch):
    # When an executable can be found
    exec = Mock()
    monkeypatch.setattr(shell, "get_command", Mock(return_value=exec))
    _, opts = pre_commit.find_executable({}, {})
    # then pre-commit should no be added to venv_install
    assert "pre-commit" not in opts.get("venv_install", [])
    # and the command should be stored in opts
    assert opts[pre_commit.CMD_OPT] == exec

    # When an executable can not be found
    monkeypatch.setattr(shell, "get_command", Mock(return_value=None))
    _, opts = pre_commit.find_executable({}, {})
    # then pre-commit should be added to venv_install
    assert "pre-commit" in opts.get("venv_install", [])
    # and the command should not be stored in opts
    assert pre_commit.CMD_OPT not in opts


def test_install(monkeypatch, caplog):
    caplog.set_level(logging.WARNING)
    # When an executable can be found
    exec = Mock()
    monkeypatch.setattr(shell, "get_command", Mock(return_value=exec))
    pre_commit.install({}, {})
    # then `pre-commit` install should run
    assert exec.called
    args, _ = exec.call_args
    assert "install" in args

    # When no executable can be found
    monkeypatch.setattr(shell, "get_command", Mock(return_value=None))
    pre_commit.install({}, {})
    # then the proper log message should be displayed
    msg = pre_commit.INSTALL_MSG.format(project_path="PROJECT_DIR")
    assert_in_logs(caplog, msg)

    # When an error occurs during installation
    caplog.set_level(logging.ERROR)
    exec = Mock(side_effect=shell.ShellCommandException)
    monkeypatch.setattr(shell, "get_command", Mock(return_value=exec))
    # then Snek should not stop, only log the error.
    pre_commit.install({}, {})
    assert_in_logs(caplog, pre_commit.ERROR_MSG)

    # When a command is available in opts
    cmd = Mock()
    exec = Mock()
    monkeypatch.setattr(shell, "get_command", Mock(return_value=exec))
    # then it should be ussed, and get_command not called
    pre_commit.install({}, {pre_commit.CMD_OPT: cmd})
    assert cmd.called
    args, _ = cmd.call_args
    assert "install" in args
    assert not exec.called


def test_add_instructions():
    old_text = get_template("readme")
    opts = {"title": "proj", "name": "proj", "description": "desc", "version": "99.9"}
    new_text, _ = pre_commit.add_instructions(opts, old_text, Mock())
    note = pre_commit.README_NOTE.format(**opts)
    assert note in new_text


# ---- Integration tests ----


def test_create_project_with_pre_commit(tmpfolder, caplog):
    caplog.set_level(logging.WARNING)
    # Given options with the pre-commit extension,
    opts = dict(project_path="proj", extensions=[pre_commit.PreCommit("pre-commit")])

    # when the project is created,
    create_project(opts)

    # then pre-commit files should exist
    assert Path("proj/.pre-commit-config.yaml").exists()
    assert Path("proj/.isort.cfg").exists()
    note = pre_commit.README_NOTE.format(name="proj")
    assert note in Path("proj/README.rst").read_text()

    # and the user should be instructed to update pre-commit
    assert_in_logs(caplog, pre_commit.UPDATE_MSG)


def test_create_project_without_pre_commit(tmpfolder):
    # Given options without the pre-commit extension,
    opts = dict(project_path="proj")

    # when the project is created,
    create_project(opts)

    # then pre-commit files should not exist
    assert not Path("proj/.pre-commit-config.yaml").exists()
    assert not Path("proj/.isort.cfg").exists()


def test_cli_with_pre_commit(tmpfolder):
    # Given the command line with the pre-commit option,
    sys.argv = ["snek", "--pre-commit", "proj"]

    # when snek runs,
    run()

    # then pre-commit files should exist
    assert Path("proj/.pre-commit-config.yaml").exists()
    assert Path("proj/.isort.cfg").exists()


def test_cli_with_pre_commit_or_pretend(tmpfolder):
    # Given the command line with the pre-commit option and pretend
    sys.argv = ["snek", "--pretend", "--pre-commit", "proj"]

    # when snek runs,
    run()

    # then pre-commit files should not exist (or the project itself)
    assert not Path("proj/.pre-commit-config.yaml").exists()
    assert not Path("proj").exists()


def test_cli_without_pre_commit(tmpfolder):
    # Given the command line without the pre-commit option,
    sys.argv = ["snek", "proj"]

    # when snek runs,
    run()

    # then pre-commit files should not exist
    assert not Path("proj/.pre-commit-config.yaml").exists()
    assert not Path("proj/.isort.cfg").exists()



"""Runner module for demoapp_data"""

import argparse
import os
import sys
from difflib import unified_diff
from pkgutil import get_data

if sys.version_info[:2] >= (3, 7):
    # TODO: Import directly (no need for conditional) when `python_requires = >= 3.7`
    from importlib.resources import read_text
else:
    from importlib_resources import read_text

from . import __name__ as pkg_name
from . import __version__ as pkg_version
from . import data as data_pkg


def get_hello_world_pkgutil():
    return get_data(pkg_name, os.path.join("data", "hello_world.txt")).decode().strip()


def get_hello_world_importlib():
    return read_text(data_pkg.__name__, "hello_world.txt").strip()


def parse_args(args):
    """
    Parse command line parameters

    :param args: command line parameters as list of strings
    :return: command line parameters as :obj:`argparse.Namespace`
    """
    parser = argparse.ArgumentParser(
        description="A demo application for Snek's unit testing"
    )
    parser.add_argument(
        "-v", "--version", action="version", version=f"demoapp_data {pkg_version}"
    )
    opts = parser.parse_args(args)
    return opts


def main(args):
    parse_args(args)
    # check several ways of reading in data
    data_pkgutil = get_hello_world_pkgutil()
    data_importlib = get_hello_world_importlib()
    diff = unified_diff(
        (data_pkgutil + "\n").splitlines(keepends=True),
        (data_importlib + "\n").splitlines(keepends=True),
        fromfile="pkgutil_data",
        tofile="importlib_data",
    )
    print(data_pkgutil + "\n-------------------------------------\n")
    msg = f"data obtained via pkgutil and importlib differ:\n\n{''.join(diff)}\n"
    print(f"data_pkgutil = {type(data_pkgutil).__name__}({data_pkgutil!r})")
    print(f"data_importlib = {type(data_importlib).__name__}({data_importlib!r})\n")
    assert data_pkgutil == data_importlib, msg


def run():
    """
    Entry point for setup.py
    """
    main(sys.argv[1:])


if __name__ == "__main__":
    main(sys.argv[1:])



import pytest

from .helpers import find_package_bin


@pytest.fixture
def tox():
    return find_package_bin("tox")


@pytest.fixture
def pre_commit():
    return find_package_bin("pre_commit", "pre-commit")


@pytest.fixture
def putup():
    return find_package_bin("snek.cli", "putup")



"""
Extension that generates configuration and script files for GitLab CI.
"""
from argparse import ArgumentParser
from typing import List

from .. import structure
from ..actions import Action, ActionParams, ScaffoldOpts, Structure
from ..operations import no_overwrite
from ..templates import get_template
from . import Extension, include
from .pre_commit import PreCommit


class GitLab(Extension):
    """Generate GitLab CI configuration files"""

    def augment_cli(self, parser: ArgumentParser):
        """Augments the command-line interface parser.
        See :obj:`~snek.extension.Extension.augment_cli`.
        """
        parser.add_argument(
            self.flag, help=self.help_text, nargs=0, action=include(PreCommit(), self)
        )
        return self

    def activate(self, actions: List[Action]) -> List[Action]:
        """Activate extension, see :obj:`~snek.extension.Extension.activate`."""
        return self.register(actions, add_files, after="define_structure")


def add_files(struct: Structure, opts: ScaffoldOpts) -> ActionParams:
    """Add .gitlab-ci.yml file to structure

    Args:
        struct: project representation as (possibly) nested :obj:`dict`.
        opts: given options, see :obj:`create_project` for an extensive list.

    Returns:
        struct, opts: updated project representation and options
    """
    files: ScaffoldOpts = {
        ".gitlab-ci.yml": (get_template("gitlab_ci"), no_overwrite())
    }

    return structure.merge(struct, files), opts



"""
Extension that generates configuration files for Yelp `pre-commit`_.

.. _pre-commit: https://pre-commit.com
"""
from functools import partial
from typing import List

from .. import shell, structure
from ..actions import Action, ActionParams, ScaffoldOpts, Structure
from ..exceptions import ShellCommandException
from ..file_system import chdir
from ..log import logger
from ..operations import FileOp, no_overwrite
from ..structure import AbstractContent, ResolvedLeaf
from ..templates import get_template
from . import Extension, venv

EXECUTABLE = "pre-commit"
CMD_OPT = "____command-pre_commit"  # we don't want this to be persisted
INSERT_AFTER = ".. _snek-notes:\n"

UPDATE_MSG = """
It is a good idea to update the hooks to the latest version:

    pre-commit autoupdate

Don't forget to tell your contributors to also install and use pre-commit.
"""

SUCCESS_MSG = "\nA pre-commit hook was installed in your repo." + UPDATE_MSG

ERROR_MSG = """
Impossible to install pre-commit automatically, please open an issue in
https://github.com/snek/snek/issues.
"""

INSTALL_MSG = f"""
A `.pre-commit-config.yaml` file was generated inside your project but in
order to make sure the hooks will run, please install `pre-commit`:

    cd {{project_path}}
    # it is a good idea to create and activate a virtualenv here
    pip install pre-commit
{UPDATE_MSG}
"""

README_NOTE = f"""
Making Changes & Contributing
=============================

This project uses `pre-commit`_, please make sure to install it before making any
changes::

    pip install pre-commit
    cd {{name}}
    pre-commit install
{UPDATE_MSG.replace(':', '::')}
.. _pre-commit: https://pre-commit.com/
"""


class PreCommit(Extension):
    """Generate pre-commit configuration file"""

    def activate(self, actions: List[Action]) -> List[Action]:
        """Activate extension

        Args:
            actions (list): list of actions to perform

        Returns:
            list: updated list of actions
        """
        actions = self.register(actions, add_files, after="define_structure")
        actions = self.register(actions, find_executable, after="define_structure")
        return self.register(actions, install, before="report_done")


def add_files(struct: Structure, opts: ScaffoldOpts) -> ActionParams:
    """Add .pre-commit-config.yaml file to structure

    Since the default template uses isort, this function also provides an
    initial version of .isort.cfg that can be extended by the user
    (it contains some useful skips, e.g. tox and venv)
    """
    files: Structure = {
        ".pre-commit-config.yaml": (get_template("pre-commit-config"), no_overwrite()),
        ".isort.cfg": (get_template("isort_cfg"), no_overwrite()),
    }

    struct = structure.modify(struct, "README.rst", partial(add_instructions, opts))
    return structure.merge(struct, files), opts


def find_executable(struct: Structure, opts: ScaffoldOpts) -> ActionParams:
    """Find the pre-commit executable that should run later in the next action...
    Or take advantage of the venv to install it...
    """
    pre_commit = shell.get_command(EXECUTABLE)
    opts = opts.copy()
    if pre_commit:
        opts.setdefault(CMD_OPT, pre_commit)
    else:
        # We can try to add it for venv to install... it will only work if the user is
        # already creating a venv anyway.
        opts.setdefault("venv_install", []).extend(["pre-commit"])

    return struct, opts


def install(struct: Structure, opts: ScaffoldOpts) -> ActionParams:
    """Attempts to install pre-commit in the project"""
    project_path = opts.get("project_path", "PROJECT_DIR")
    pre_commit = opts.get(CMD_OPT) or shell.get_command(EXECUTABLE, venv.get_path(opts))
    # ^  try again after venv, maybe it was installed
    if pre_commit:
        try:
            with chdir(opts.get("project_path", "."), **opts):
                pre_commit("install", pretend=opts.get("pretend"))
            logger.warning(SUCCESS_MSG)
            return struct, opts
        except ShellCommandException:
            logger.error(ERROR_MSG, exc_info=True)

    logger.warning(INSTALL_MSG.format(project_path=project_path))
    return struct, opts


def add_instructions(
    opts: ScaffoldOpts, content: AbstractContent, file_op: FileOp
) -> ResolvedLeaf:
    """Add pre-commit instructions to README"""
    text = structure.reify_content(content, opts)
    if text is not None:
        i = text.find(INSERT_AFTER)
        assert i > 0, f"{INSERT_AFTER!r} not found in README template:\n{text}"
        j = i + len(INSERT_AFTER)
        text = text[:j] + README_NOTE.format(**opts) + text[j:]
    return text, file_op



"""Internal library for manipulating package dependencies and requirements."""

import re
from itertools import chain
from typing import Iterable, List

from packaging.requirements import InvalidRequirement, Requirement

# setuptools version is now enforced via `install_requires`

BUILD = ("setuptools_scm>=5",)
"""Dependencies that will be required to build the created project"""
RUNTIME = ('importlib-metadata; python_version<"3.8"',)
# TODO: Remove `importlib-metadata` when `python_requires = >= 3.8`
"""Dependencies that will be required at runtime by the created project"""
ISOLATED = ("setuptools>=46.1.0", "setuptools_scm[toml]>=5", *BUILD[1:])
"""Dependencies for isolated builds (:pep:`517`/:pep:`518`).
- setuptools min version might be slightly higher then the version required at runtime.
- setuptools_scm requires an optional dependency to work with pyproject.toml
"""
# Although version 36.6.0 introduces PEP517 implementation,
# version 46.1.0 fix a bug with setuptools.finalize_distribution_options,
# which is a hook used by setuptools_scm (better safe then sorry).

REQ_SPLITTER = re.compile(r";(?!\s*(python|platform|implementation|os|sys)_)", re.M)
"""Regex to split requirements that considers both `setup.cfg specs`_ and :pep:`508`
(in a *good enough* simplified fashion).

.. _setup.cfg specs: https://setuptools.pypa.io/en/latest/userguide/declarative_config.html
"""  # noqa


def split(requirements: str) -> List[str]:
    """Split a combined requirement string (such as the values for ``setup_requires``
    and ``install_requires`` in ``setup.cfg``) into a list of individual requirement
    strings, that can be used in :obj:`is_included`, :obj:`get_requirements_str`,
    :obj:`remove`, etc...
    """
    lines = requirements.splitlines()
    deps = (dep.strip() for line in lines for dep in REQ_SPLITTER.split(line) if dep)
    return [dep for dep in deps if dep]  # Remove empty deps


def deduplicate(requirements: Iterable[str]) -> List[str]:
    """Given a sequence of individual requirement strings, e.g.
    ``["platformdirs>=1.4.4", "packaging>20.0"]``, remove the duplicated packages.
    If a package is duplicated, the last occurrence stays.
    """
    return list({attempt_pkg_name(r): r for r in requirements}.values())


def remove(requirements: Iterable[str], to_remove: Iterable[str]) -> List[str]:
    """Given a list of individual requirement strings, e.g. ``["platformdirs>=1.4.4",
    "packaging>20.0"]``, remove the requirements in ``to_remove``.
    """
    removable = {attempt_pkg_name(r) for r in to_remove}
    return [r for r in requirements if attempt_pkg_name(r) not in removable]


def add(requirements: Iterable[str], to_add: Iterable[str] = BUILD) -> List[str]:
    """Given a sequence of individual requirement strings, add ``to_add`` to it.
    By default adds :obj:`BUILD` if ``to_add`` is not given."""
    return deduplicate(chain(requirements, to_add))


def attempt_pkg_name(requirement: str) -> str:
    """In the case the given string is a dependency specification (:pep:`508`/
    :pep`440`), it returns the "package name" part of dependency (without versions).
    Otherwise, it returns the same string (removed the comment marks).
    """
    req = requirement.strip("#").strip()
    try:
        return Requirement(req).name
    except InvalidRequirement:
        return req



from unittest.mock import Mock

import pytest

from snek.exceptions import InvalidIdentifier
from snek.identification import (
    dasherize,
    deterministic_name,
    deterministic_sort,
    get_id,
    is_valid_identifier,
    levenshtein,
    make_valid_identifier,
    underscore,
)


def test_is_valid_identifier():
    bad_names = [
        "has whitespace",
        "has-hyphen",
        "has_special_char$",
        "1starts_with_digit",
    ]
    for bad_name in bad_names:
        assert not is_valid_identifier(bad_name)
    valid_names = ["normal_variable_name", "_private_var", "_with_number1"]
    for valid_name in valid_names:
        assert is_valid_identifier(valid_name)


def test_make_valid_identifier():
    assert make_valid_identifier("has whitespaces ") == "has_whitespaces"
    assert make_valid_identifier("has-hyphon") == "has_hyphon"
    assert make_valid_identifier("special chars%") == "special_chars"
    assert make_valid_identifier("UpperCase") == "uppercase"
    with pytest.raises(InvalidIdentifier):
        make_valid_identifier("def")


def test_levenshtein():
    s1 = "born"
    s2 = "burn"
    assert levenshtein(s1, s2) == 1
    s2 = "burnt"
    assert levenshtein(s1, s2) == 2
    assert levenshtein(s2, s1) == 2
    s2 = ""
    assert levenshtein(s2, s1) == 4


def test_dasherize():
    assert dasherize("hello_world") == "hello-world"
    assert dasherize("helloworld") == "helloworld"
    assert dasherize("") == ""


def test_underscore():
    assert underscore("HelloWorld") == "hello_world"
    assert underscore("Hello-World") == "hello_world"
    assert underscore("Hello   World") == "hello_world"
    assert underscore("Hello---World") == "hello_world"
    assert underscore("HelLo-WorLd") == "hel_lo_wor_ld"
    assert underscore("helloworld") == "helloworld"
    assert underscore("PYTHON") == "p_y_t_h_o_n"
    assert underscore("") == ""


@pytest.mark.parametrize(
    "obj, name",
    [
        (underscore, "snek.identification.underscore"),
        (None, "...NoneType"),
        (InvalidIdentifier(), "snek.exceptions.InvalidIdentifier"),
    ],
)
def test_deterministic_name(obj, name):
    assert deterministic_name(obj) == name


def test_deterministic_sort():
    ex = InvalidIdentifier()
    ext1 = Mock(__module__="snek.extension", __qualname__="Extension")
    ext2 = Mock(__module__="snekext", __qualname__="Extension")
    assert deterministic_sort([underscore, None, ext2, ext1, ex]) == [
        None,
        ex,
        ext1,
        underscore,
        ext2,
    ]


def test_get_id():
    def custom_action(structure, options):
        return structure, options

    custom_action.__module__ = "awesome_module"
    assert get_id(custom_action) == "awesome_module:custom_action"



from pathlib import Path

import pytest

from snek.actions import discover, get_default_options
from snek.actions import init_git as orig_init_git
from snek.actions import register, unregister, verify_project_dir
from snek.api import bootstrap_options
from snek.exceptions import (
    ActionNotFound,
    DirectoryAlreadyExists,
    DirectoryDoesNotExist,
    GitNotConfigured,
    GitNotInstalled,
    NestedRepository,
)
from snek.structure import define_structure


def test_discover():
    # Given an extension with actions,
    def fake_action(struct, opts):
        return struct, opts

    def extension(actions):
        return [fake_action] + actions

    # When discover is called,
    actions = discover([extension])

    # Then the extension actions should be listed alongside default actions.
    assert get_default_options in actions
    assert fake_action in actions


def test_get_default_opts():
    opts = bootstrap_options(project_path="project", package="package")
    _, opts = get_default_options({}, opts)
    assert all(k in opts for k in "project_path update force author".split())
    assert isinstance(opts["extensions"], list)
    assert isinstance(opts["requirements"], list)


def test_get_default_opts_with_nogit(nogit_mock):
    with pytest.raises(GitNotInstalled):
        get_default_options({}, dict(project_path="my-project"))


def test_get_default_opts_with_git_not_configured(noconfgit_mock):
    with pytest.raises(GitNotConfigured):
        get_default_options({}, dict(project_path="my-project"))


def test_verify_project_dir_when_project_doesnt_exist_and_updating(tmpfolder, git_mock):
    opts = dict(project_path=Path("my-project"), update=True)
    with pytest.raises(DirectoryDoesNotExist):
        verify_project_dir({}, opts)


def test_verify_project_dir_when_project_exist_but_not_updating(tmpfolder, git_mock):
    tmpfolder.ensure("my-project", dir=True)
    opts = dict(project_path=Path("my-project"), update=False, force=False)
    with pytest.raises(DirectoryAlreadyExists):
        verify_project_dir({}, opts)


def test_verify_nested_repository(tmpfolder, git_mock):
    Path(tmpfolder, ".git").mkdir()  # Pretend parent is a git repo using git_mock
    opts = dict(project_path=Path(tmpfolder, "my-project"), update=False, force=False)
    with pytest.raises(NestedRepository):
        verify_project_dir({}, opts)


def test_verify_nested_repository_when_forcing(tmpfolder, git_mock):
    Path(tmpfolder, ".git").mkdir()  # Pretend parent is a git repo using git_mock
    opts = dict(project_path=Path(tmpfolder, "my-project"), update=False, force=True)
    # No error should be raised if `force` is True
    verify_project_dir({}, opts)


def custom_action(structure, options):
    return structure, options


custom_action.__module__ = "awesome_module"


def init_git(structure, options):
    """Fake action that shares the same name as a default action."""
    return structure, options


def test_register_before():
    # Given an action list,
    pipeline = [orig_init_git]
    # When a new action is registered before another, using the function name
    # as position reference,
    pipeline = register(pipeline, custom_action, before="init_git")
    # Then this action should be correctly placed
    assert pipeline == [custom_action, orig_init_git]


def test_register_after():
    # Given an action list,
    pipeline = [orig_init_git]
    # When a new action is registered after another, using the function name
    # as position reference,
    pipeline = register(pipeline, custom_action, after="init_git")
    # Then this action should be correctly placed
    assert pipeline == [orig_init_git, custom_action]


def test_register_with_qualified_name():
    # Given an action list with actions that share the same name,
    pipeline = [orig_init_git, init_git]
    # When a new action is registered using the "qualified" name
    # (module+function) as position reference,
    position = "snek.actions:init_git"
    pipeline = register(pipeline, custom_action, after=position)
    # Then this action should be correctly placed
    assert pipeline == [orig_init_git, custom_action, init_git]


def test_register_default_position():
    # Given an action list with define_structure,
    pipeline = [orig_init_git, define_structure, init_git]
    # When a new action is registered without position reference,
    pipeline = register(pipeline, custom_action)
    # Then this action should be placed after define_structure
    assert pipeline == [orig_init_git, define_structure, custom_action, init_git]


def test_register_with_invalid_reference():
    # Given an action list,
    pipeline = [orig_init_git]
    # When a new action is registered using an invalid reference,
    with pytest.raises(ActionNotFound):
        # Then the proper exception should be raised,
        pipeline = register(pipeline, custom_action, after="undefined_action")
    # And the action list should remain the same
    assert pipeline == [orig_init_git]


def test_unregister():
    # Given an action list with name conflict,
    pipeline = [custom_action, init_git, orig_init_git]
    # When an action is unregistered by name,
    pipeline = unregister(pipeline, "init_git")
    # Then the first match should be removed
    assert pipeline == [custom_action, orig_init_git]


def test_unregister_with_qualified_name():
    # Given an action list with name conflict,
    pipeline = [custom_action, init_git, orig_init_git]
    # When an action is unregistered by "qualified" name,
    pipeline = unregister(pipeline, "snek.actions:init_git")
    # Then the correct match should be removed
    assert pipeline == [custom_action, init_git]


def test_unregister_with_undefined_action():
    # Given an action list,
    pipeline = [orig_init_git]
    # When a undefined action is unregistered,
    with pytest.raises(ActionNotFound):
        # Then the proper exception should be raised,
        pipeline = unregister(pipeline, "undefined_action")
    # And the action list should remain the same
    assert pipeline == [orig_init_git]



import argparse
import sys

import pytest

from snek import extensions
from snek.exceptions import ErrorLoadingExtension

from .extensions import __name__ as test_extensions_pkg
from .extensions.helpers import make_extension

if sys.version_info[:2] >= (3, 8):
    # TODO: Import directly (no need for conditional) when `python_requires = >= 3.8`
    from importlib.metadata import EntryPoint  # pragma: no cover
else:
    from importlib_metadata import EntryPoint  # pragma: no cover


def test_extension():
    parser = argparse.ArgumentParser()
    extension = make_extension("MyExtension")
    extension.augment_cli(parser)
    opts = vars(parser.parse_args(["--my-extension"]))
    assert opts["extensions"] == [extension]


def test_extension_append():
    parser = argparse.ArgumentParser()
    extension1 = make_extension("MyExtension1")
    extension2 = make_extension("MyExtension2")
    parser.set_defaults(extensions=[extension1])

    extension2.augment_cli(parser)
    opts = vars(parser.parse_args(["--my-extension2"]))
    assert opts["extensions"] == [extension1, extension2]


def test_include():
    parser = argparse.ArgumentParser()
    my_extensions = [make_extension(f"MyExtension{n}") for n in range(7)]
    parser.add_argument("--opt", nargs=0, action=extensions.include(*my_extensions))
    opts = vars(parser.parse_args(["--opt"]))
    assert opts["extensions"] == my_extensions


def test_store_with():
    parser = argparse.ArgumentParser()
    my_extensions = [make_extension(f"MyExtension{n}") for n in range(7)]
    parser.add_argument("--opt", action=extensions.store_with(*my_extensions))
    opts = vars(parser.parse_args(["--opt", "42"]))
    assert opts["extensions"] == my_extensions
    assert opts["opt"] == "42"


def test_store_with_type():
    parser = argparse.ArgumentParser()
    my_extensions = [make_extension(f"MyExtension{n}") for n in range(7)]
    parser.add_argument("--opt", type=int, action=extensions.store_with(*my_extensions))
    opts = vars(parser.parse_args(["--opt", "42"]))
    assert opts["extensions"] == my_extensions
    assert opts["opt"] == 42


def test_load_from_entry_point__error():
    # This module does not exist, so Python will have some trouble loading it
    # EntryPoint(name, value, group)
    fake = EntryPoint("fake", "snekext.SOOOOO___fake___:Fake", "snek.cli")
    with pytest.raises(ErrorLoadingExtension):
        extensions.load_from_entry_point(fake)


def test_load_from_entry_point__old_api():
    # The following module/class exists but uses an old version of the extensions API
    # therefore, we should have a meaningful error when trying to load it.
    entry = f"{test_extensions_pkg}.incompatible_v3_api_fake_extension:FakeExtension"
    fake = EntryPoint("fake", entry, "snek.cli")
    with pytest.raises(ErrorLoadingExtension):
        extensions.load_from_entry_point(fake)


def test_iterate_entry_points():
    ext_iter = extensions.iterate_entry_points()
    assert hasattr(ext_iter, "__iter__")
    ext_list = list(ext_iter)
    name_list = [e.name for e in ext_list]
    for ext in ("cirrus", "pre_commit", "no_cli", "namespace", "venv"):
        assert ext in name_list


def test_list_from_entry_points():
    # Should return a list with all the extensions registered in the entrypoints
    ext_list = extensions.list_from_entry_points()
    orig_len = len(ext_list)
    assert all(isinstance(e, extensions.Extension) for e in ext_list)
    name_list = [e.name for e in ext_list]
    for ext in ("cirrus", "pre_commit", "no_cli", "namespace", "venv"):
        assert ext in name_list

    # a filtering function can be passed to avoid loading extensions that are not needed
    ext_list = extensions.list_from_entry_points(filtering=lambda e: e.name != "cirrus")
    name_list = [e.name for e in ext_list]
    assert len(ext_list) == orig_len - 1
    assert "cirrus" not in name_list



"""Unit tests of everything related to retrieving the version

There are four tree states we want to check:
 A: sitting on the 1.0 tag
 B: dirtying the tree after 1.0
 C: a commit after a tag, clean tree
 D: a commit after a tag, dirty tree

Each test will run inside a different venv in a temporary directory, so they
can execute in parallel and not interfere with each other.
"""

import json
import os
import shutil
from contextlib import contextmanager
from datetime import datetime, timezone
from pathlib import Path
from shutil import copyfile

import pytest

from snek import shell
from snek.cli import main as putup
from snek.extensions import venv
from snek.file_system import chdir
from snek.shell import git

__location__ = Path(__file__).parent


pytestmark = pytest.mark.slow


@pytest.fixture
def demoapp(tmpfolder):
    app = DemoApp(tmpfolder)
    with chdir(app.pkg_path):
        yield app


@pytest.fixture
def demoapp_data(tmpfolder):
    app = DemoApp(tmpfolder, data=True)
    with chdir(app.pkg_path):
        yield app


class DemoApp:
    def __init__(self, tmpdir, data=None):
        self.name = "demoapp"
        if data:
            self.name += "_data"
        self.pkg_path = Path(str(tmpdir), self.name)
        self.built = False
        self.installed = False

        self.venv_path = Path(str(tmpdir), ".venv")
        venv.create(self.venv_path)
        assert self.venv_path.exists()

        self._cmd_opts = {
            "shell": os.name == "nt",  # needed on Windows to pass env vars correctly.
            "include_path": False,  # avoid leaking executables outside the venv
        }
        self.python = shell.get_command("python", self.venv_path, **self._cmd_opts)
        # ^ Python inside the venv...
        #   Different from ``shell.python`` which is the one running the tests (.tox)
        self._cli = None

        self.data = data
        self.dist = None

        with chdir(str(tmpdir)):
            self._generate()

    def _generate(self):
        putup([self.name])
        with chdir(self.name):
            demoapp_src_dir = __location__ / self.name
            demoapp_dst_root = self.pkg_path
            demoapp_dst_pkg = demoapp_dst_root / "src" / self.name
            copyfile(demoapp_src_dir / "runner.py", demoapp_dst_pkg / "runner.py")
            git("add", demoapp_dst_pkg / "runner.py")
            for file in "setup.cfg setup.py pyproject.toml".split():
                copyfile(demoapp_src_dir / file, demoapp_dst_root / file)
                git("add", demoapp_dst_root / file)
            if self.data:
                data_src_dir = demoapp_src_dir / "data"
                data_dst_dir = demoapp_dst_pkg / "data"
                os.mkdir(data_dst_dir)
                pkg_file = data_dst_dir / "__init__.py"
                pkg_file.write_text("")
                git("add", pkg_file)
                for file in "hello_world.txt".split():
                    copyfile(data_src_dir / file, data_dst_dir / file)
                    git("add", data_dst_dir / file)
            git("commit", "-m", "Added basic application logic")

        # setuptools-scm is used for tests
        if os.name == "os":
            # Windows lacks some certificates
            self.pip("install", "-q", "certifi", "setuptools_scm")
        else:
            self.pip("install", "-q", "setuptools_scm")

    def pip(self, *cmd, **kwargs):
        return self.python("-m", "pip", *cmd, **kwargs)

    def check_not_installed(self):
        installed = [
            e["name"] for e in json.loads(list(self.pip("list", "--format", "json"))[0])
        ]
        dirty = [self.name, "UNKNOWN"]
        app_list = [x for x in dirty if x in installed]
        if app_list:
            msg = f"Dirty virtual environment:\n{', '.join(app_list)} found"
            raise RuntimeError(msg)

    @contextmanager
    def guard(self, attr):
        if getattr(self, attr):
            raise RuntimeError("For simplicity, just build/install once per package")
        yield
        setattr(self, attr, True)

    @property
    def cli(self):
        if self._cli is None:
            self._cli = shell.get_command(self.name, self.venv_path, **self._cmd_opts)
        return self._cli

    def get_version(self):
        out = "".join(self.cli("--version"))
        return out.replace(self.name, "").strip()

    def build(self, dist="wheel", cli_opts=()):
        with self.guard("built"), chdir(self.pkg_path):
            # For the sake of speed, we use the same Python running the tests
            # (inside .tox) and skip isolation (no extra venv just for building)
            args = [self.pkg_path, *cli_opts]
            shell.python("-m", "build", "--no-isolation", f"--{dist}", *args)
        self.dist = dist
        return self

    @property
    def dist_file(self):
        return next((self.pkg_path / "dist").glob(self.name + "*"))

    def install(self, edit=False):
        with self.guard("installed"), chdir(self.pkg_path):
            self.check_not_installed()
            if edit or self.dist is None:
                self.pip("install", "-e", ".")
            else:
                self.pip("install", self.dist_file)
        return self

    def installed_path(self):
        if not self.installed:
            return None

        cmd = f"import {self.name}; print({self.name}.__path__[0])"
        return Path(next(self.python("-c", cmd)))

    def make_dirty_tree(self):
        dirty_file = self.pkg_path / "src" / self.name / "runner.py"
        with open(dirty_file, "a") as fh:
            fh.write("\n\ndirty_variable = 69\n")
        return self

    def make_commit(self):
        with chdir(self.pkg_path):
            git("commit", "-a", "-m", "message")
        return self

    def rm_git_tree(self):
        git_path = self.pkg_path / ".git"
        shutil.rmtree(git_path)
        return self

    def tag(self, name, message):
        with chdir(self.pkg_path):
            git("tag", "-a", name, "-m", message)
        return self


def check_version(version, exp_version, dirty=False):
    today = datetime.now(timezone.utc).strftime("%Y%m%d")
    dirty_tag = f".d{today}"
    # ^  this depends on the local strategy configured for setuptools_scm...
    #    the default 'node-and-date'

    sep = "+" if "+" in version else "_"
    # ^  for some setuptools version a directory with + is generated, sometimes _

    if dirty:
        ver, local = version.split(sep)
        assert local.endswith(dirty_tag)
        assert ver == exp_version
    else:
        ver = version.split(sep)
        if len(ver) > 1:
            assert not ver[1].endswith(dirty_tag)
        assert ver[0] == exp_version


def test_sdist_install(demoapp):
    (demoapp.build("sdist").install())
    exp = "0.0.post1.dev2"
    check_version(demoapp.get_version(), exp, dirty=False)


def test_sdist_install_dirty(demoapp):
    (
        demoapp.tag("v0.1", "first release")
        .make_dirty_tree()
        .make_commit()
        .make_dirty_tree()
        .build("sdist")
        .install()
    )
    exp = "0.1.post1.dev1"
    check_version(demoapp.get_version(), exp, dirty=True)


def test_sdist_install_with_1_0_tag(demoapp):
    (
        demoapp.make_dirty_tree()
        .make_commit()
        .tag("v1.0", "final release")
        .build("sdist")
        .install()
    )
    exp = "1.0"
    check_version(demoapp.get_version(), exp, dirty=False)


def test_sdist_install_with_1_0_tag_dirty(demoapp):
    demoapp.tag("v1.0", "final release").make_dirty_tree().build("sdist").install()
    exp = "1.0.post1.dev0"
    check_version(demoapp.get_version(), exp, dirty=True)


def test_wheel_install(demoapp):
    demoapp.build("wheel").install()
    exp = "0.0.post1.dev2"
    check_version(demoapp.get_version(), exp, dirty=False)


def test_sdist_install_with_data(demoapp_data):
    demoapp_data.build("sdist").install()
    out = "".join(demoapp_data.cli())
    exp = "Hello World"
    assert out.startswith(exp)


def test_wheel_install_with_data(demoapp_data):
    demoapp_data.build("wheel").install()
    path = demoapp_data.installed_path()
    assert path.exists()
    assert (path / "data/__init__.py").exists()
    assert (path / "data/hello_world.txt").exists()
    assert (path / "runner.py").exists()
    out = "".join(demoapp_data.cli())
    exp = "Hello World"
    assert out.startswith(exp)


def test_edit_install_with_data(demoapp_data):
    demoapp_data.install(edit=True)
    out = "".join(demoapp_data.cli())
    exp = "Hello World"
    assert out.startswith(exp)



"""This file is a fixture for TEST PURPOSES ONLY and uses an OUTDATED version of the
extensions API, and therefore will result in errors when imported
"""

from snek.api import Extension, helpers

# The module helpers was removed in version 4
# The class Extension was moved to the snek.extensions module


class FakeExtension(Extension):
    """Help text on commandline when running putup -h"""

    def activate(self, actions):
        """Activate extension

        Args:
            actions (list): list of actions to perform

        Returns:
            list: updated list of actions
        """
        actions = helpers.register(actions, self.action, after="create_structure")
        return actions

    def action(self, struct, opts):
        """Perform some actions that modifies the structure and options

        Args:
            struct (dict): project representation as (possibly) nested
                :obj:`dict`.
            opts (dict): given options, see :obj:`create_project` for
                an extensive list.

        Returns:
            struct, opts: updated project representation and options
        """
        print(self)
        return struct, opts



"""
Shell commands like git, django-admin etc.
"""

import functools
import os
import shlex
import shutil
import subprocess
import sys
from functools import lru_cache
from pathlib import Path
from typing import Callable, Dict, Iterable, Iterator, List, Optional, Union

from .exceptions import ShellCommandException
from .log import logger

PathLike = Union[str, os.PathLike]

IS_POSIX = os.name == "posix"
IS_WINDOWS = sys.platform == "win32"

# The following default flags were borrowed from Github's docs:
# https://docs.github.com/en/github/getting-started-with-github/associating-text-editors-with-git
EDITORS: Dict[str, List[str]] = {
    # -- $EDITOR and $VISUAL should be considered first --
    # Terminal based
    "sensible-editor": [],  # default editor in Debian-based systems like Ubuntu
    "nvim": [],  # if a person has nvim installed, chances are they like it as EDITOR
    "nano": [],  # beginner-friendly, vim users will likely have EDITOR=vim already set
    "vim": [],
    # GUI
    "subl": ["-w"],
    "code": ["--wait"],
    "mate": ["-w"],  # OS X specific
    "atom": ["--wait"],
    # Fallbacks (we tried reasonably hard to find a good editor...)
    "notepad": [],  # Windows
    "vi": [],  # POSIX
}
"""Programs to be tried (in sequence) when calling :obj:`edit` and :obj:`get_editor` in
the case the environment variables EDITOR and VISUAL are not set.
"""


_GIT_CMD = "git"
_GIT_CMD_WIN = "git.exe"


class ShellCommand:
    """Shell command that can be called with flags like git('add', 'file')

    Args:
        command: command to handle
        shell: run the command in the shell (``True`` by default).
        cwd: current working dir to run the command

    The produced command can be called with the following keyword arguments:

        - **pretend** (*bool*): skip execution (but log) when pretending.
          ``False`` by default.

    The positional arguments are passed to the underlying shell command.
    In the case the path to the executable contains spaces of any other special shell
    character, ``command`` needs to be properly quoted.
    """

    def __init__(
        self,
        command: str,
        shell: bool = True,
        cwd: Optional[str] = None,
    ):
        self._command = command
        self._shell = shell
        self._cwd = cwd

    def run(self, *args, **kwargs) -> subprocess.CompletedProcess:
        """Execute command with the given arguments via :obj:`subprocess.run`."""
        command = f"{self._command} {join(args)}".strip()
        logger.report("run", command, context=self._cwd)

        should_pretend = kwargs.pop("pretend", False)
        if should_pretend:
            return subprocess.CompletedProcess(command, 0, None, None)

        opts: dict = {
            "shell": self._shell,
            "cwd": self._cwd,
            "stdout": subprocess.PIPE,
            "stderr": subprocess.STDOUT,
            "universal_newlines": True,
            "env": {
                **os.environ,
                # try to disable i18n
                **dict(LC_ALL="C", LANGUAGE=""),
            },
            **kwargs,  # allow overwriting defaults
        }
        if self._shell:
            return subprocess.run(command, **opts)
            # ^ `check_output` does not seem to support terminal editors
        return subprocess.run(shlex.split(command, posix=IS_POSIX), **opts)
        # ^ Joining and then splitting is the only way of supporting all the cases

    def __call__(self, *args, **kwargs) -> Iterator[str]:
        """Execute the command, returning an iterator for the resulting text output"""
        try:
            completed = self.run(*args, **kwargs)
        except FileNotFoundError as e:
            logger.report("info", f'last command failed with "{e!s}"')
            raise ShellCommandException(str(e)) from e

        try:
            completed.check_returncode()
        except subprocess.CalledProcessError as e:
            stdout, stderr = (e or "" for e in (completed.stdout, completed.stderr))
            stdout, stderr = (e.strip() for e in (stdout, stderr))
            sep = "; " if stdout and stderr else ""
            msg = sep.join([stdout, stderr])
            logger.report("info", f'last command failed with "{msg}"')
            raise ShellCommandException(msg) from e

        return (line for line in (completed.stdout or "").splitlines())


def shell_command_error2exit_decorator(func: Callable):
    """Decorator to convert given ShellCommandException to an exit message

    This avoids displaying nasty stack traces to end-users
    """

    @functools.wraps(func)
    def func_wrapper(*args, **kwargs):
        try:
            func(*args, **kwargs)
        except ShellCommandException as e:
            cause = e.__cause__
            reason = cause.output if cause and hasattr(cause, "output") else e
            print(f"{e}:\n{reason}")
            sys.exit(1)

    return func_wrapper


# ToDo: Change this to just `cache` from Python 3.9 on.
@lru_cache(maxsize=None)
def get_git_cmd(**args) -> ShellCommand:
    """Retrieve the git shell command depending on the current platform

    Args:
        **args: additional keyword arguments to :obj:`~.ShellCommand`
    """
    if IS_WINDOWS:  # pragma: no cover
        # ^  CI setup does not aggregate Windows coverage
        for shell in (True, False):
            cmd = ShellCommand(_GIT_CMD_WIN, shell=shell, **args)
            try:
                cmd("--version")
            except ShellCommandException:
                continue
            return cmd  # available and works with either shell=True or shell=False
        return cmd  # not available, but we return it anyway for better error messages
    return ShellCommand(_GIT_CMD, **args)


def command_exists(cmd: str) -> bool:
    """Check if command exists

    Args:
        cmd: executable name
    """
    return shutil.which(cmd) is not None


def get_executable(
    name: str, prefix: PathLike = sys.prefix, include_path=True
) -> Optional[str]:
    """Find an executable in the system, if available.

    Args:
        name: name of the executable
        prefix: look on this directory, exclusively or in addition to $PATH
            depending on the value of ``include_path``. Defaults to :obj:`sys.prefix`.
        include_path: when True the functions tries to look in the entire $PATH.


    Note:
        The return value might contain whitespaces. If this value is used in a shell
        environment, it needs to be quote properly to avoid the underlying shell
        interpreter splitting the executable path.
    """
    executable = shutil.which(name)
    if include_path and executable:
        return executable

    candidates = list(Path(prefix).resolve().glob(f"*/{name}*"))
    # ^  this works in virtual envs and both Windows and POSIX
    if candidates:
        path = (f.parent for f in sorted(candidates, key=lambda p: len(str(p))))
        return shutil.which(name, path=os.pathsep.join(str(p) for p in path))
        # ^  which will guarantee we find an executable and not only a regular file

    return None


def get_command(
    name: str, prefix: PathLike = sys.prefix, include_path=True, shell=True, **kwargs
) -> Optional[ShellCommand]:
    """Similar to :obj:`get_executable` but return an instance of :obj:`ShellCommand`
    if it is there to be found.
    Additional kwargs will be passed to the :obj:`ShellCommand` constructor.
    """
    executable = get_executable(name, prefix, include_path)
    if not executable:
        return None

    if shell:
        executable = join([executable])

    kwargs["shell"] = shell
    return ShellCommand(executable, **kwargs)


def get_editor(**kwargs) -> str:
    """Get an available text editor program"""
    from_env = os.getenv("VISUAL") or os.getenv("EDITOR")
    if from_env:
        return from_env  # user is responsible for proper quoting

    candidates = ((get_executable(e), opts) for e, opts in EDITORS.items())
    editor, opts = next((c for c in candidates if c[0]), (None, [""]))
    if editor:
        return join([editor, *opts])

    msg = "No text editor found in your system, please set EDITOR in your environment"
    raise ShellCommandException(msg)


def edit(file: PathLike, *args, **kwargs) -> Path:
    """Open a text editor and returns back a :obj:`Path` to file, after user editing."""
    editor = ShellCommand(get_editor())
    editor(file, *args, **{"stdout": None, "stderr": None, **kwargs})
    # ^  stdout/stderr=None => required for a terminal editor to open properly
    return Path(file)


def join(parts: Iterable[Union[str, PathLike]]) -> str:
    """Join different parts of a shell command into a string, quoting whitespaces."""
    if IS_WINDOWS:  # pragma: no cover
        # ^  CI setup does not aggregate Windows coverage
        return subprocess.list2cmdline(map(str, parts))

    return " ".join(shlex.quote(str(p)) for p in parts)
    # ^  TODO: Replace with `shlex.join(map(str, parts))` when Python >= 3.8


def git(*args, **kwargs) -> Iterator[str]:
    """Command for git"""
    return get_git_cmd()(*args, **kwargs)  # delayed, so errors show up with --verbose


#: Command for python
python = ShellCommand(sys.executable)



import logging
import os
import sys
from unittest.mock import Mock

import pytest

from snek import cli
from snek.exceptions import ErrorLoadingExtension
from snek.file_system import localize_path as lp

from .log_helpers import find_report

if sys.version_info[:2] >= (3, 8):
    # TODO: Import directly (no need for conditional) when `python_requires = >= 3.8`
    from importlib.metadata import EntryPoint  # pragma: no cover
else:
    from importlib_metadata import EntryPoint  # pragma: no cover


def test_parse_args():
    args = ["my-project"]
    opts = cli.parse_args(args)
    assert opts["project_path"] == "my-project"


def test_parse_verbose_option():
    for quiet in ("--verbose", "-v"):
        args = ["my-project", quiet]
        opts = cli.parse_args(args)
        assert opts["log_level"] == logging.INFO


def test_parse_default_log_level():
    args = ["my-project"]
    opts = cli.parse_args(args)
    assert opts["log_level"] == logging.WARNING


def test_parse_pretend():
    for flag in ["--pretend", "-P"]:
        opts = cli.parse_args(["my-project", flag])
        assert opts["pretend"]
    opts = cli.parse_args(["my-project"])
    assert not opts["pretend"]


def test_parse_list_actions():
    opts = cli.parse_args(["my-project", "--list-actions"])
    assert opts["command"] == cli.list_actions
    opts = cli.parse_args(["my-project"])
    assert opts["command"] == cli.run_scaffold


def test_parse_license_finds_best_fit():
    examples = {
        "apache": "Apache-2.0",
        "artistic": "Artistic-2.0",
        "affero": "AGPL-3.0-only",
        "eclipse": "EPL-1.0",
        "new-bsd": "BSD-3-Clause",
        "mozilla": "MPL-2.0",
        "gpl3": "GPL-3.0-only",
    }

    for key, value in examples.items():
        opts = cli.parse_args(["my-project", "--license", key])
        assert opts["license"] == value

    # option not passed
    assert "license" not in cli.parse_args(["my_project"])


def test_verbose_main(tmpfolder, git_mock, caplog):
    args = ["my-project", "--verbose"]
    cli.main(args)
    assert os.path.exists(args[0])

    # Check for some log messages
    assert find_report(caplog, "invoke", "get_default_options")
    assert find_report(caplog, "invoke", "verify_options_consistency")
    assert find_report(caplog, "invoke", "define_structure")
    assert find_report(caplog, "invoke", "create_structure")
    assert find_report(caplog, "create", "setup.py")
    assert find_report(caplog, "create", lp("my_project/__init__.py"))
    assert find_report(caplog, "run", "git init")
    assert find_report(caplog, "run", "git add")


def test_pretend_main(tmpfolder, git_mock, caplog):
    for flag in ["--pretend", "-P"]:
        args = ["my-project", flag]
        cli.main(args)
        assert not os.path.exists(args[0])

        # Check for some log messages
        assert find_report(caplog, "invoke", "get_default_options")
        assert find_report(caplog, "invoke", "verify_options_consistency")
        assert find_report(caplog, "invoke", "define_structure")
        assert find_report(caplog, "invoke", "create_structure")
        assert find_report(caplog, "create", "setup.py")
        assert find_report(caplog, "create", lp("my_project/__init__.py"))
        assert find_report(caplog, "run", "git init")
        assert find_report(caplog, "run", "git add")


def test_main_when_updating(tmpfolder, capsys, git_mock):
    args = ["my-project"]
    cli.main(args)
    args = ["--update", "my-project"]
    cli.main(args)
    assert os.path.exists(args[1])
    out, _ = capsys.readouterr()
    assert "Update accomplished!" in out


def test_main_with_list_actions(tmpfolder, capsys, isolated_logger):
    # When putup is called with --list-actions,
    args = ["my-project", "--no-tox", "--list-actions"]
    cli.main(args)
    # then the action list should be printed,
    out, _ = capsys.readouterr()
    assert "Planned Actions" in out
    assert "snek.actions:get_default_options" in out
    assert "snek.structure:define_structure" in out
    assert "snek.extensions.no_tox:remove_files" in out
    assert "snek.structure:create_structure" in out
    assert "snek.actions:init_git" in out
    # but no project should be created
    assert not os.path.exists(args[0])


def test_wrong_extension(monkeypatch, tmpfolder):
    # Given an entry point with some problems is registered in the snek.cli group
    # (e.g. failing implementation, wrong dependencies that cause the python file to
    # fail to evaluate)
    fake = EntryPoint("fake", "snekext.SOOO__fake__:Fake", "snek.cli")
    entry_points_mock = Mock(return_value={"snek.cli": [fake]})
    monkeypatch.setattr("snek.extensions.entry_points", entry_points_mock)
    with pytest.raises(ErrorLoadingExtension, match=r".*error loading.*fake.*"):
        # When putup is called with the corresponding flag
        args = ["my-project"]
        cli.main(args)
        entry_points_mock.assert_called()
    # Then the CLI should display a meaningful error message


def test_run(tmpfolder, git_mock):
    sys.argv = ["snek", "my-project"]
    cli.run()
    assert os.path.exists(sys.argv[1])


def test_get_log_level():
    assert cli.get_log_level([]) == logging.WARNING
    assert cli.get_log_level(["--pretend"]) == logging.INFO
    assert cli.get_log_level(["--verbose"]) == logging.INFO
    assert cli.get_log_level(["--very-verbose"]) == logging.DEBUG

    # Make sure it also works with sys.argv
    sys.argv = ["putup", "--very-verbose"]
    assert cli.get_log_level() == logging.DEBUG



import getpass
import os
import socket
from pathlib import Path
from unittest.mock import MagicMock as Mock

import pytest
from configupdater import ConfigUpdater

from snek import actions, cli, exceptions, info, repo, structure, templates


def test_username_with_git(git_mock):
    username = info.username()
    assert isinstance(username, str)
    assert len(username) > 0


def test_username_with_no_git(nogit_mock):
    username = info.username()
    assert isinstance(username, str)
    assert getpass.getuser() == username


def test_username_error(git_mock, monkeypatch):
    fake_git = Mock(side_effect=exceptions.ShellCommandException)
    monkeypatch.setattr(info.shell, "git", fake_git)
    # on windows getpass might fail
    monkeypatch.setattr(info.getpass, "getuser", Mock(side_effect=SystemError))
    with pytest.raises(exceptions.GitNotConfigured):
        info.username()


def test_email_with_git(git_mock):
    email = info.email()
    assert "@" in email


def test_email_with_nogit(nogit_mock):
    email = info.email()
    assert socket.gethostname() == email.split("@")[1]


def test_email_error(git_mock, monkeypatch):
    fake_git = Mock(side_effect=exceptions.ShellCommandException)
    monkeypatch.setattr(info.shell, "git", fake_git)
    # on windows getpass might fail
    monkeypatch.setattr(info.getpass, "getuser", Mock(side_effect=SystemError))
    with pytest.raises(exceptions.GitNotConfigured):
        info.email()


def test_git_is_installed(git_mock):
    assert info.is_git_installed()


def test_git_is_wrongely_installed(nogit_mock):
    assert not info.is_git_installed()


def test_git_is_not_installed(nogit_cmd_mock):
    assert not info.is_git_installed()


def test_is_git_configured(git_mock):
    assert info.is_git_configured()


def test_git_is_configured_via_env_vars(monkeypatch):
    monkeypatch.setenv("GIT_AUTHOR_NAME", "John Doe")
    monkeypatch.setenv("GIT_AUTHOR_EMAIL", "johndoe@email.com")
    assert info.is_git_configured()


def test_is_git_not_configured(noconfgit_mock):
    assert not info.is_git_configured()


def test_check_git_not_installed(nogit_cmd_mock):
    with pytest.raises(exceptions.GitNotInstalled):
        info.check_git()


def test_check_git_not_configured(noconfgit_mock):
    with pytest.raises(exceptions.GitNotConfigured):
        info.check_git()


def test_check_git_installed_and_configured(git_mock):
    info.check_git()


def test_project_without_args(tmpfolder):
    old_args = [
        "my_project",
        "-u",
        "http://www.blue-yonder.com/",
        "-d",
        "my description",
    ]
    cli.main(old_args)
    args = ["my_project"]
    opts = cli.parse_args(args)
    new_opts = info.project(opts)
    assert new_opts["url"] == "http://www.blue-yonder.com/"
    assert new_opts["package"] == "my_project"
    assert new_opts["license"] == "MIT"
    assert new_opts["description"] == "my description"


def test_project_with_args(tmpfolder):
    old_args = [
        "my_project",
        "-u",
        "http://www.blue-yonder.com/",
        "-d",
        "my description",
    ]
    cli.main(old_args)
    args = ["my_project", "-u", "http://www.google.com/", "-d", "other description"]
    opts = cli.parse_args(args)
    new_opts = info.project(opts)
    assert new_opts["url"] == "http://www.google.com/"
    assert new_opts["package"] == "my_project"
    assert new_opts["description"] == "other description"


def test_project_with_no_setup(tmpfolder):
    os.mkdir("my_project")
    args = ["my_project"]
    opts = cli.parse_args(args)
    with pytest.raises(FileNotFoundError):
        info.project(opts)


def test_project_with_wrong_setup(tmpfolder):
    os.mkdir("my_project")
    open("my_project/setup.py", "a").close()
    args = ["my_project"]
    opts = cli.parse_args(args)
    with pytest.raises(FileNotFoundError):
        info.project(opts)


def test_project_old_setupcfg(tmpfolder):
    demoapp = Path(__file__).parent / "demoapp"
    with pytest.raises(exceptions.SnekTooOld):
        info.project({}, config_path=demoapp)


def test_project_extensions_not_found(tmpfolder):
    _, opts = actions.get_default_options({}, {})
    cfg = ConfigUpdater().read_string(templates.setup_cfg(opts))
    cfg["snek"]["extensions"] = "x_foo_bar_x"
    (tmpfolder / "setup.cfg").write(str(cfg))
    with pytest.raises(exceptions.ExtensionNotFound) as exc:
        info.project(opts)
    assert "x_foo_bar_x" in str(exc.value)


@pytest.mark.no_fake_config_dir
def test_config_dir_error(monkeypatch):
    # no_fake_config_dir => avoid previous mock of config_dir

    # If for some reason something goes wrong when trying to find the config dir
    user_config_dir_mock = Mock(side_effect=SystemError)
    monkeypatch.setattr(info.platformdirs, "user_config_dir", user_config_dir_mock)
    # And no default value is given
    # Then an error should be raised
    with pytest.raises(exceptions.ImpossibleToFindConfigDir):
        print("config_dir", info.config_dir())
        user_config_dir_mock.assert_called_once()


def test_config_file_default(monkeypatch):
    # When config_dir does not find the correct config directory
    monkeypatch.setattr(info, "config_dir", Mock(return_value=None))
    # And there are a default file
    demoapp_setup = Path(__file__).parent / "demoapp" / "setup.cfg"
    # Then the default file should be returned
    assert info.config_file(default=demoapp_setup) == demoapp_setup


def test_best_fit_license():
    # If the name matches => return the name
    for license in templates.licenses.keys():
        assert info.best_fit_license(license) == license
    # Lower case
    assert info.best_fit_license("0bsd") == "0BSD"
    # No dashes
    assert info.best_fit_license("mpl2") == "MPL-2.0"
    assert info.best_fit_license("gpl2") == "GPL-2.0-only"
    assert info.best_fit_license("gpl3") == "GPL-3.0-only"
    assert info.best_fit_license("bsd2") == "BSD-2-Clause"
    assert info.best_fit_license("bsd3") == "BSD-3-Clause"
    # Popular nicknames
    assert info.best_fit_license("apache") == "Apache-2.0"
    assert info.best_fit_license("artistic") == "Artistic-2.0"
    assert info.best_fit_license("affero") == "AGPL-3.0-only"
    assert info.best_fit_license("eclipse") == "EPL-1.0"
    assert info.best_fit_license("bsd0") == "0BSD"
    assert info.best_fit_license("new-bsd") == "BSD-3-Clause"
    assert info.best_fit_license("simple-bsd") == "BSD-2-Clause"
    assert info.best_fit_license("cc0") == "CC0-1.0"
    assert info.best_fit_license("none") == "Proprietary"
    assert info.best_fit_license("public-domain") == "Unlicense"
    # Or later vs only
    assert info.best_fit_license("gpl3-only") == "GPL-3.0-only"
    assert info.best_fit_license("gpl2-later") == "GPL-2.0-or-later"
    # Default
    assert info.best_fit_license("") == "MIT"


def test_dirty_workspace(tmpfolder):
    project = "my_project"
    struct = {"dummyfile": "NO CONTENT"}
    structure.create_structure(struct, dict(project_path=project))
    repo.init_commit_repo(project, struct)
    path = tmpfolder / project
    assert info.is_git_workspace_clean(path)
    with open(str(path / "dummyfile"), "w") as fh:
        fh.write("CHANGED\n")
    assert not info.is_git_workspace_clean(path)



"""
Provide general information about the system, user and the package itself.
"""

import copy
import getpass
import os
import socket
import sys
from enum import Enum
from operator import itemgetter
from pathlib import Path
from typing import Optional, Set, cast, overload

import platformdirs
from configupdater import ConfigUpdater
from packaging.version import Version

from . import __name__ as PKG_NAME
from . import shell, toml
from .exceptions import (
    ExtensionNotFound,
    GitNotConfigured,
    GitNotInstalled,
    ImpossibleToFindConfigDir,
    SnekTooOld,
    ShellCommandException,
)
from .file_system import PathLike, chdir
from .identification import deterministic_sort, levenshtein, underscore
from .log import logger
from .templates import ScaffoldOpts, licenses, parse_extensions

CONFIG_FILE = "default.cfg"
"""Snek's own config file name"""

PYPROJECT_TOML: PathLike = "pyproject.toml"
SETUP_CFG: PathLike = "setup.cfg"


class GitEnv(Enum):
    author_name = "GIT_AUTHOR_NAME"
    author_email = "GIT_AUTHOR_EMAIL"
    author_date = "GIT_AUTHOR_DATE"
    committer_name = "GIT_COMMITTER_NAME"
    committer_email = "GIT_COMMITTER_EMAIL"
    committer_date = "GIT_COMMITTER_DATE"


def username() -> str:
    """Retrieve the user's name"""
    user = os.getenv(GitEnv.author_name.value)
    if user is None:
        try:
            user = next(shell.git("config", "--get", "user.name"))
            user = user.strip()
        except ShellCommandException:
            try:
                # On Windows the getpass commands might fail if 'USERNAME'
                # env var is not set
                user = getpass.getuser()
            except Exception as ex:
                logger.debug("Impossible to find hostname", exc_info=True)
                raise GitNotConfigured from ex
    return user


def email() -> str:
    """Retrieve the user's email"""
    mail = os.getenv(GitEnv.author_email.value)
    if mail is None:
        try:
            mail = next(shell.git("config", "--get", "user.email"))
            mail = mail.strip()
        except ShellCommandException:
            try:
                # On Windows the getpass commands might fail
                user = getpass.getuser()
                host = socket.gethostname()
                mail = f"{user}@{host}"
            except Exception as ex:
                logger.debug("Impossible to determine email", exc_info=True)
                raise GitNotConfigured from ex
    return mail


def is_git_installed() -> bool:
    """Check if git is installed"""
    logger.report("check", "is git installed...")
    try:
        shell.git("--version")
    except ShellCommandException:
        return False
    return True


def is_git_configured() -> bool:
    """Check if user.name and user.email is set globally in git

    Check first git environment variables, then config settings.
    This will also return false if git is not available at all.

    Returns:
        True if it is set globally, False otherwise
    """
    logger.report("check", "is git configured...")
    if os.getenv(GitEnv.author_name.value) and os.getenv(GitEnv.author_email.value):
        return True
    else:
        try:
            for attr in ("name", "email"):
                shell.git("config", "--get", f"user.{attr}")
        except ShellCommandException:
            return False
        else:
            return True


def check_git():
    """Checks for git and raises appropriate exception if not

    Raises:
       :class:`~.GitNotInstalled`: when git command is not available
       :class:`~.GitNotConfigured`: when git does not know user information
    """
    if not is_git_installed():
        raise GitNotInstalled
    if not is_git_configured():
        raise GitNotConfigured


def is_git_workspace_clean(path: PathLike) -> bool:
    """Checks if git workspace is clean

    Args:
        path: path to git repository

     Raises:
        :class:`~.GitNotInstalled`: when git command is not available
        :class:`~.GitNotConfigured`: when git does not know user information
    """
    logger.report("check", "is workspace of git clean...")
    check_git()
    try:
        with chdir(path):
            shell.git("diff-index", "--quiet", "HEAD", "--")
    except ShellCommandException:
        return False
    return True


def project(
    opts: ScaffoldOpts,
    config_path: Optional[PathLike] = None,
    config_file: Optional[PathLike] = None,
) -> ScaffoldOpts:
    """Update user options with the options of an existing config file

    Args:
        opts: options of the project

        config_path: path where config file can be found
            (default: ``opts["project_path"]``)

        config_file: if ``config_path`` is a directory, name of the config file,
            relative to it (default: ``setup.cfg``)

    Returns:
        Options with updated values

    Raises:
        :class:`~.SnekTooOld`: when Snek is to old to update from
        :class:`~.NoSnekProject`: when project was not generated with Snek
    """
    # Lazily load the following function to avoid circular dependencies
    from .extensions import NO_LONGER_NEEDED  # TODO: NO_LONGER_SUPPORTED
    from .extensions import list_from_entry_points as list_extensions

    opts = copy.deepcopy({k: v for k, v in opts.items() if not callable(v)})
    # ^  functions/lambdas are not deepcopy-able

    path = config_path or cast(PathLike, opts.get("project_path", "."))

    cfg = read_setupcfg(path, config_file).to_dict()
    if "snek" not in cfg:
        raise SnekTooOld

    snek = cfg.pop("snek", {})
    metadata = cfg.pop("metadata", {})

    license = metadata.get("license")
    existing = {
        "package": snek.pop("package", None),
        "name": metadata.get("name"),
        "author": metadata.get("author"),
        "email": metadata.get("author_email") or metadata.get("author-email"),
        "url": metadata.get("url"),
        "description": cast(str, metadata.get("description", "")).strip(),
        "license": license and best_fit_license(license),
    }
    existing = {k: v for k, v in existing.items() if v}  # Filter out non stored values

    # Overwrite only if user has not provided corresponding cli argument
    # Derived/computed parameters should be set by `get_default_options`
    opts = {**existing, **opts}

    # Complement the cli extensions with the ones from configuration
    not_found_ext: Set[str] = set()
    if "extensions" in snek:
        cfg_extensions = parse_extensions(snek.pop("extensions", None) or "")
        opt_extensions = {ext.name for ext in opts.setdefault("extensions", [])}
        add_extensions = cfg_extensions - opt_extensions

        other_ext = list_extensions(filtering=lambda e: e.name in add_extensions)
        not_found_ext = add_extensions - {e.name for e in other_ext} - NO_LONGER_NEEDED
        opts["extensions"] = deterministic_sort(opts["extensions"] + other_ext)

    if not_found_ext:
        raise ExtensionNotFound(list(not_found_ext))

    # TODO: NO_LONGER_SUPPORTED => raise Exception(use older Snek)

    # The remaining values in the snek section can be added to opts
    # if not specified yet. Useful when extensions define other options.
    for key, value in snek.items():
        opts.setdefault(key, value)

    return opts


def best_fit_license(txt: Optional[str]) -> str:
    """Finds proper license name for the license defined in txt"""
    corresponding = {
        **{v.replace("license_", ""): k for k, v in licenses.items()},
        **{_simplify_license_name(k): k for k in licenses},
        **{k: k for k in licenses},  # last defined: possibly overwrite
    }
    lic = underscore(txt or list(licenses)[0]).replace("_", "")
    candidates = {underscore(k).replace("_", ""): v for k, v in corresponding.items()}
    ratings = {k: levenshtein(lic.lower(), k.lower()) for k, v in candidates.items()}
    return candidates[min(ratings.items(), key=itemgetter(1))[0]]


def _simplify_license_name(name: str) -> str:
    for term in ("-Clause", "-or-later", "-only"):
        name = name.replace(term, "")
    return name


def read_setupcfg(path: PathLike, filename=SETUP_CFG) -> ConfigUpdater:
    """Reads-in a configuration file that follows a setup.cfg format.
    Useful for retrieving stored information (e.g. during updates)

    Args:
        path: path where to find the config file
        filename: if ``path`` is a directory, ``name`` will be considered a file
            relative to ``path`` to read (default: setup.cfg)

    Returns:
        Object that can be used to read/edit configuration parameters.
    """
    path = Path(path)
    if path.is_dir():
        path = path / (filename or SETUP_CFG)

    updater = ConfigUpdater()
    updater.read(str(path), encoding="utf-8")

    logger.report("read", path)

    return updater


def read_pyproject(path: PathLike, filename=PYPROJECT_TOML) -> toml.TOMLMapping:
    """Reads-in a configuration file that follows a pyproject.toml format.

    Args:
        path: path where to find the config file
        filename: if ``path`` is a directory, ``name`` will be considered a file
            relative to ``path`` to read (default: setup.cfg)

    Returns:
        Object that can be used to read/edit configuration parameters.
    """
    file = Path(path)
    if file.is_dir():
        file = file / (filename or PYPROJECT_TOML)

    config = toml.loads(file.read_text(encoding="utf-8"))
    logger.report("read", file)
    return config


def get_curr_version(project_path: PathLike):
    """Retrieves the Snek version that put up the scaffold

    Args:
        project_path: path to project

    Returns:
        Version: version specifier
    """
    setupcfg = read_setupcfg(project_path).to_dict()
    return Version(str(setupcfg["snek"]["version"]))


(RAISE_EXCEPTION,) = list(Enum("default", "RAISE_EXCEPTION"))  # type: ignore
"""When no default value is passed, an exception should be raised"""


@overload
def config_dir(prog: str = PKG_NAME, org: Optional[str] = None) -> Path:
    ...


@overload
def config_dir(
    prog: str = PKG_NAME,
    org: Optional[str] = None,
    default: Optional[Path] = RAISE_EXCEPTION,
) -> Optional[Path]:
    ...


def config_dir(prog=PKG_NAME, org=None, default=RAISE_EXCEPTION):
    """Finds the correct place where to read/write configurations for the given app.

    Args:
        prog: program name (defaults to snek)
        org: organisation/author name (defaults to the same as ``prog``)
        default: default value to return if an exception was raise while
            trying to find the config dir. If no default value is passed, an
            :obj:`~.ImpossibleToFindConfigDir` execution is raised.

    Please notice even if the directory doesn't exist, if its path is possible
    to calculate, this function will return a Path object (that can be used to
    create the directory)

    Returns:
        Location somewhere in the user's home directory where to put the configs.
    """
    try:
        return Path(platformdirs.user_config_dir(prog, org, roaming=True))
    except Exception as ex:
        if default is not RAISE_EXCEPTION:
            logger.debug("Error when finding config dir %s", ex, exc_info=True)
            return default
        raise ImpossibleToFindConfigDir() from ex


@overload
def config_file(
    name: str = CONFIG_FILE, prog: str = PKG_NAME, org: Optional[str] = None
) -> Path:
    ...


@overload
def config_file(
    name: str = CONFIG_FILE,
    prog: str = PKG_NAME,
    org: Optional[str] = None,
    default: Optional[Path] = RAISE_EXCEPTION,
) -> Optional[Path]:
    ...


def config_file(name=CONFIG_FILE, prog=PKG_NAME, org=None, default=RAISE_EXCEPTION):
    """Finds a file inside :obj:`config_dir`.

    Args:
        name: file you are looking for

    The other args are the same as in :obj:`config_dir` and have the same
    meaning.

    Returns:
        Location of the config file or default if an error happened.
    """
    default_file = default
    if default is not RAISE_EXCEPTION:
        default = None

    dir = config_dir(prog, org, default)
    if dir is None:
        return default_file

    return dir / name


def _old_macos_config_dir(new_dir: Path) -> Path:
    return Path(
        str(new_dir)
        .replace(os.sep, "/")
        .replace("/Library/Application Support/", "/Library/Preferences/")
    )


def _migrate_old_macos_config(prog: str = PKG_NAME, org: Optional[str] = None):
    """Compensate for macOS backward incompatible change in platformdirs 3.0.0"""
    if not sys.platform.startswith("darwin"):
        return

    try:
        new_dir = config_dir(prog, org)
        old_dir = _old_macos_config_dir(new_dir)
        if new_dir.exists() or not old_dir.exists():
            return

        logger.report("move", str(old_dir), target=str(new_dir))
        new_dir.parent.mkdir(parents=True, exist_ok=True)
        old_dir.rename(new_dir)
    except Exception:  # pragma: no cover
        logger.debug(
            "Error trying to migrate old macOS config dir. "
            "If you have an older Snek config file, please make sure to "
            f"manually copy it to `~/Library/Application Support/{prog}/`",
            exc_info=True,
        )



"""
External API for accessing Snek programmatically via Python.
"""
from enum import Enum
from functools import reduce
from pathlib import Path

from . import __version__ as VERSION
from . import actions, info
from .exceptions import DirectErrorForUser, NoSnekProject

# -------- Options --------

(NO_CONFIG,) = list(Enum("ConfigFiles", "NO_CONFIG"))  # type: ignore
"""This constant is used to tell Snek to not load any extra configuration file,
not even the default ones
Usage::

    create_project(opts, config_files=NO_CONFIG)

Please notice that the ``setup.cfg`` file inside an project being updated will
still be considered.
"""

DEFAULT_OPTIONS = {
    "update": False,
    "force": False,
    "description": "Add a short description here!",
    "url": "https://github.com/",
    "license": "MIT",
    "version": VERSION,
    "extensions": [],
    "config_files": [],  # Overloaded in bootstrap_options for lazy evaluation
}
"""Default values for Snek's options.

Options that can be derived from the values of other options (e.g. ``package`` can be
derived from ``project_path`` when not explicitly passed) are computed in
:obj:`snek.actions.get_default_options`.

When ``config_files`` is empty, a default value is computed dynamically by
:obj:`snek.info.config_file` before the start of Snek's action pipeline.

Warning:
    Default values might be dynamically overwritten by ``config_files`` or, during
    updates, existing ``setup.cfg``.
"""


def bootstrap_options(opts=None, **kwargs):
    """Internal API: augment the given options with minimal defaults
    and existing configurations saved in files (e.g. ``setup.cfg``)

    See list of arguments in :obj:`create_project`.
    Returns a dictionary of options.

    Warning:
        This function is not part of the public Python API of Snek, and therefore
        might change even in minor/patch releases (not bounded to semantic versioning).

    Note:
        This function does not replace the :obj:`snek.actions.get_default_options`
        action. Instead it is needed to ensure that action works correctly.
    """
    opts = opts.copy() if opts else {}
    opts.update(kwargs)

    # Clean up:
    opts = {k: v for k, v in opts.items() if v or v is False}
    # ^  remove empty items, so we ensure setdefault works

    # Add options stored in config files:
    info._migrate_old_macos_config()
    default_files = [info.config_file(default=None)]
    opts.setdefault("config_files", [f for f in default_files if f and f.exists()])
    # ^  make sure the file exists before passing it ahead
    opts = _read_existing_config(opts)

    # Add defaults last, so they don't overwrite:
    opts.update({k: v for k, v in DEFAULT_OPTIONS.items() if k not in opts})

    opts["version"] = VERSION  # always update version
    return opts


# -------- Public API --------


def create_project(opts=None, **kwargs):
    """Create the project's directory structure

    Args:
        opts (dict): options of the project
        **kwargs: extra options, passed as keyword arguments

    Returns:
        tuple: a tuple of `struct` and `opts` dictionary

    Valid options include:

    :Project Information:   - **project_path** (:obj:`os.PathLike` or :obj:`str`)

    :Naming:                - **name** (*str*): as in ``pip install`` or in PyPI
                            - **package** (*str*): Python identifier as in ``import``
                              (without namespace)

    :Package Information:   - **author** (*str*)
                            - **email** (*str*)
                            - **release_date** (*str*)
                            - **year** (*str*)
                            - **title** (*str*)
                            - **description** (*str*)
                            - **url** (*str*)
                            - **classifiers** (*str*)
                            - **requirements** (*list*)

    :Snek Control:    - **update** (*bool*)
                      - **force** (*bool*)
                      - **pretend** (*bool*)
                      - **extensions** (*list*)
                      - **config_files** (*list* or ``NO_CONFIG``)

    Some of these options are equivalent to the command line options, others
    are used for creating the basic python package meta information, but the
    ones in the "Snek Control" group affects how the "scaffolding" behaves.

    When the **force** flag is ``True``, existing files will be overwritten.
    When the **update** flag is ``True``, Snek will consider that some
    files can be updated (usually the packaging boilerplate),
    but will keep others intact.
    When the **pretend** flag is ``True``, the project will not be
    created/updated, but the expected outcome will be logged.

    The **extensions** list may contain any object that follows the
    :ref:`extension API <extensions>`. Note that some Snek features, such
    as cirrus, tox and pre-commit support, are implemented as built-in
    extensions. In order to use these features it is necessary to include the
    respective objects in the extension list. All built-in extensions are
    accessible via :mod:`snek.extensions` submodule.

    Finally, when ``setup.cfg``-like files are added to the **config_files** list,
    Snek will read it's options from there in addition to the ones already passed.
    If the list is empty, the default configuration file is used. To avoid reading any
    existing configuration, please pass ``config_file=NO_CONFIG``.
    See :doc:`usage </usage>` for more details.

    Note that extensions may define extra options. For example, the
    cookiecutter extension define a ``cookiecutter`` option that
    should be the address to the git repository used as template and the ``namespace``
    extension define a ``namespace`` option with the name of a PEP 420 compatible
    (and possibly nested) namespace.
    """
    opts = bootstrap_options(opts, **kwargs)
    pipeline = actions.discover(opts["extensions"])

    # call the actions to generate final struct and opts
    return reduce(actions.invoke, pipeline, ({}, opts))


# -------- Auxiliary functions (Private) --------


def _read_existing_config(opts):
    """Read existing config files first listed in ``opts["config_files"]``
    and then ``setup.cfg`` inside ``opts["project_path"]``
    """
    config_files = opts["config_files"]
    if config_files is not NO_CONFIG:
        paths = (Path(f).resolve() for f in config_files)
        deduplicated = {p: p for p in paths}
        # ^  using a dict instead of a set to preserve the order the files were given
        # ^  we do not mute errors here if the file does not exist. Let us be
        #    explicit.
        opts = reduce(info.project, deduplicated.keys(), opts)

    if opts.get("update"):
        try:
            opts = info.project(opts)
            # ^  In case of an update read and parse setup.cfg inside project
        except DirectErrorForUser:
            raise
        except Exception as e:
            raise NoSnekProject from e

    return opts



#!/usr/bin/env python
import sys
from pathlib import Path

from snek.api import create_project
from snek.cli import run
from snek.extensions.no_tox import NoTox


def test_create_project_with_tox(tmpfolder):
    # when the project is created,
    opts = dict(project_path="proj")
    create_project(opts)

    # then tox files are created by default
    assert Path("proj/tox.ini").exists()


def test_create_project_without_tox(tmpfolder):
    # when the project is created with no_tox
    opts = dict(project_path="proj", extensions=[NoTox()])
    create_project(opts)

    # then tox files should not exist
    assert not Path("proj/tox.ini").exists()


def test_cli_with_tox(tmpfolder):
    # when the project is created with the CLI
    sys.argv = ["snek", "proj"]
    run()

    # then tox files should exist
    assert Path("proj/tox.ini").exists()


def test_cli_without_tox(tmpfolder):
    # Given the command line with --no-tox
    sys.argv = ["snek", "proj", "-vv", "--no-tox"]

    # when snek runs,
    run()

    # then tox files should not exist
    assert not Path("proj/tox.ini").exists()


def test_cli_without_tox_but_pretend(tmpfolder):
    # Given the command line with --no-tox and --pretend
    sys.argv = ["snek", "proj", "-vv", "--no-tox", "--pretend"]

    # when snek runs,
    run()

    # then tox files should not exist (or even the project)
    assert not Path("proj/tox.ini").exists()
    assert not Path("proj").exists()



"""Create a virtual environment for the project"""
import argparse
from contextlib import suppress
from pathlib import Path
from typing import List, Optional

from .. import dependencies as deps
from ..actions import Action, ActionParams, ScaffoldOpts, Structure
from ..file_system import PathLike, chdir
from ..identification import get_id
from ..log import logger
from ..shell import get_command, get_executable
from . import Extension, store_with

DEFAULT: PathLike = ".venv"
"""Default directory name for collocated virtual environment that will be created"""


class Venv(Extension):
    """\
    Create a virtual environment for the project (using virtualenv or stdlib's venv).
    Default location: "{DEFAULT}".

    If ``virtualenv`` is available, it will be used, since it has some advantages over
    stdlib's ``venv`` (such as being faster, see https://virtualenv.pypa.io/en/stable/).

    Notice that even if part of Python's stdlib, ``venv`` is not guaranteed to be
    installed, some OS/distributions (such as Ubuntu) require an explicit installation.
    If you have problems, try installing virtualenv with pip and run the command again.
    """

    persist = False  # We just want the virtual env to be created on fresh projects

    def augment_cli(self, parser: argparse.ArgumentParser):
        """Augments the command-line interface parser.
        See :obj:`~snek.extension.Extension.augment_cli`.
        """
        parser.add_argument(
            self.flag,
            action=store_with(self),
            nargs="?",
            const=DEFAULT,
            default=argparse.SUPPRESS,
            type=Path,
            help=self.help_text.format(DEFAULT=DEFAULT),
        )
        parser.add_argument(
            "--venv-install",
            action=store_with(self),
            nargs="+",
            default=argparse.SUPPRESS,
            metavar="PACKAGE",
            help="install packages inside the created venv. "
            "The packages can have dependency ranges as if they would be written to a "
            "`requirements.txt` file, but remember to use quotes to avoid messing with "
            "the terminal",
        )
        return self

    def activate(self, actions: List[Action]) -> List[Action]:
        """Activate extension, see :obj:`~snek.extension.Extension.activate`."""
        actions = self.register(actions, run, after="init_git")
        actions = self.register(actions, install_packages, after=get_id(run))
        return self.register(actions, instruct_user, before="report_done")


def run(struct: Structure, opts: ScaffoldOpts) -> ActionParams:
    """Action that will create a virtualenv for the project"""

    project_path = opts["project_path"]
    venv_path = Path(opts.get("venv", DEFAULT))

    with chdir(project_path, **opts):
        if venv_path.is_dir():
            logger.report("skip", venv_path)
            return struct, _fix_opts(opts)

        create(venv_path, opts.get("pretend"))

    return struct, _fix_opts(opts)


def install_packages(struct: Structure, opts: ScaffoldOpts) -> ActionParams:
    """Install the specified packages inside the created venv."""

    packages = opts.get("venv_install")
    if not packages:
        return struct, opts

    pretend = opts.get("pretend")
    venv_path = get_path(opts)

    if not pretend:
        pip = get_command("pip", venv_path, include_path=False)
        if not pip:
            raise NotInstalled(f"pip cannot be found inside {venv_path}")
        pip("install", "-U", *deps.deduplicate(packages))

    logger.report("run", f"pip install -U {' '.join(packages)} [{venv_path}]")
    return struct, opts


def instruct_user(struct: Structure, opts: ScaffoldOpts) -> ActionParams:
    """Simply display a message reminding the user to activate the venv."""

    venv = opts.get("venv", DEFAULT)
    if opts.get("pretend"):
        logger.warning(f"\nA virtual environment was created: `{venv}`.\n")
        return struct, opts

    project = Path(opts["project_path"]).resolve()
    with chdir(project, **opts):
        venv_path = Path(venv)
        python_exe = get_executable("python", venv_path, include_path=False)
        pip_exe = get_executable("pip", venv_path, include_path=False)

    if python_exe and pip_exe:
        python = Path(python_exe).relative_to(project)
        pip = Path(pip_exe).relative_to(project)

        logger.warning(
            f"\nA virtual environment was created in the `{venv_path}` directory.\n"
            "You need to activate it (every time you open a new terminal),\n"
            f"or call directly {python} and {pip}.\n"
        )

    return struct, opts


def get_path(opts: ScaffoldOpts, default=DEFAULT) -> Path:
    """Get the path to the venv that will be created."""
    with chdir(opts.get("project_path", "."), **opts):
        return Path(opts.get("venv", default)).resolve()


def create_with_virtualenv(path: Path, pretend=False):
    import virtualenv

    args = [str(path)]

    if pretend:
        virtualenv.session_via_cli(args)
    else:
        logger.warning("\nInstalling virtual environment, it might take a while...\n")
        virtualenv.cli_run(args)

    logger.report("virtualenv", path)


def create_with_stdlib(path: Path, pretend=False):
    import venv

    if not pretend:
        logger.warning("\nInstalling virtual environment, it might take a while...\n")
        venv.create(str(path), with_pip=True)

    logger.report("venv", path)


def create(path: Path, pretend=False):
    """Create the virtual environment with the first technique available.
    (``virtualenv`` is preferred because it is faster).
    """
    for creator in (create_with_virtualenv, create_with_stdlib):
        with suppress(ImportError):
            creator(path, pretend)
            break
    else:
        # no break statement found, so no creator function executed correctly
        raise NotInstalled()


class NotInstalled(ImportError):
    """Neither virtualenv or venv are installed in the computer. Please check the
    following alternatives:

    - ``virtualenv`` can be installed via pip
    - ``venv`` is supposed to installed by default with Python3, however some OS or
      distributions (such as Ubuntu) break the standard library in a series of packages
      that need to be manually installed via OS package manager.
      You can try to search for a ``python-venv``, ``python3-venv`` or similar in the
      official repositories.
    """

    def __init__(self, msg: Optional[str] = None):
        super().__init__(msg or self.__doc__)


def _fix_opts(opts: ScaffoldOpts) -> ScaffoldOpts:
    pkgs = opts.get("venv_install")
    if not pkgs:
        return opts

    return {**opts, "venv_install": deps.split(pkgs) if isinstance(pkgs, str) else pkgs}



from argparse import ArgumentError
from inspect import cleandoc
from itertools import chain, product
from os import environ
from pathlib import Path
from unittest.mock import Mock

import pytest

from snek import api, cli
from snek import file_system as fs
from snek.extensions import venv

from ..helpers import ArgumentParser, disable_import, rmpath, uniqstr

# ---- "Isolated" tests ----


def parse(*args, set_defaults=None):
    parser = ArgumentParser()
    parser.set_defaults(**(set_defaults or {}))
    venv.Venv().augment_cli(parser)
    return vars(parser.parse_args(args))


def test_cli_opts():
    # no opts
    opts = parse()
    assert "venv" not in opts
    assert "venv_install" not in opts
    assert not opts.get("extensions")
    # opt but no value
    opts = parse("--venv")
    assert opts["venv"] == Path(".venv")
    assert [e.name for e in opts["extensions"]] == ["venv"]
    # opt and value
    opts = parse("--venv", ".here")
    assert opts["venv"] == Path(".here")
    # venv-install
    opts = parse("--venv-install", "platformdirs>=1.1,<2", "six")
    opts = parse("--venv-install", "platformdirs>=1.1,<2", "six")
    assert opts["venv_install"] == ["platformdirs>=1.1,<2", "six"]
    assert [e.name for e in opts["extensions"]] == ["venv"]
    # venv-install but no value
    with pytest.raises((ArgumentError, TypeError, SystemExit)):
        # ^  TypeError happens because argparse tries to iterate over the --config opts
        #    since it is marked with nargs='+'
        #    ArgumentError and SystemExit might happen depending on the version
        #    of Python when there is a parse error.
        opts = parse("--venv-install")


def test_with_virtualenv_available(monkeypatch, tmpfolder):
    # When virtualenv is available
    virtualenv_mock = Mock()
    monkeypatch.setattr(venv, "create_with_virtualenv", virtualenv_mock)
    venv.run({}, {"project_path": Path(tmpfolder), "venv": ".venv"})
    # it will be called
    virtualenv_mock.assert_called_once()


@disable_import("virtualenv")
def test_without_virtualenv_available(monkeypatch, tmpfolder):
    # When virtualenv is not available
    venv_mock = Mock()
    monkeypatch.setattr(venv, "create_with_stdlib", venv_mock)
    venv.run({}, {"project_path": Path(tmpfolder), "venv": ".venv"})
    # venv will be called
    venv_mock.assert_called_once()


@disable_import("virtualenv")
@disable_import("venv")
def test_with_none_available(tmpfolder):
    # When virtualenv and venv are not available
    # then an exception will be raised
    with pytest.raises(venv.NotInstalled):
        venv.run({}, {"project_path": Path(tmpfolder), "venv": ".venv"})


def test_already_exists(monkeypatch, tmpfolder):
    virtualenv_mock = Mock()
    venv_mock = Mock()
    monkeypatch.setattr(venv, "create_with_virtualenv", virtualenv_mock)
    monkeypatch.setattr(venv, "create_with_stdlib", venv_mock)
    # When the venv directory already exist
    (Path(tmpfolder) / ".venv").mkdir()
    venv.run({}, {"project_path": Path(tmpfolder), "venv": ".venv"})
    # It should skip
    virtualenv_mock.assert_not_called()
    venv_mock.assert_not_called()


# ---- Integration tests ----


@pytest.mark.slow
@pytest.mark.parametrize(
    "creator, pretend",
    product((venv.create_with_virtualenv, venv.create_with_stdlib), (True, False)),
)
def test_creators(tmp_path_factory, creator, pretend):
    folder = tmp_path_factory.mktemp(f"test_creators_{uniqstr()}_")
    with fs.chdir(folder):  # ensure parametrized tests do not share folders
        path = Path(".venv")
        try:
            creator(path, pretend=pretend)
        except Exception:
            if environ.get("USING_CONDA") == "true":
                pytest.skip(
                    "Creating venvs more than one level deep inside conda is tricky"
                    "and error prone. Here we use at least 2 levels (tox > venv)."
                )
            else:
                raise

        if pretend:
            assert not path.exists()
        else:
            assert path.is_dir()
            assert list(path.glob("*/python*"))
            assert list(chain(path.glob("*/pip*"), path.glob("*/activate*")))

    rmpath(folder)


@pytest.mark.slow
def test_install_packages(existing_venv, tmpfolder):
    venv_path = existing_venv.path.resolve()
    tmp = Path(str(tmpfolder)).resolve()

    # Given packages are not installed
    for pkg in "pytest pip-compile".split():
        assert venv.get_executable(pkg, venv_path, False) is None

    # when we run install_packages
    opts = {
        "project_path": tmp,
        "venv": venv_path,
        "venv_install": ["pytest>=6.0.0", "pip-tools"],
    }
    venv.install_packages({}, opts)

    # then they should be installed
    for pkg in "pytest pip-compile".split():
        bin_dir = str(Path(venv.get_executable(pkg, venv_path, False)).parent)
        assert bin_dir.lower().startswith(str(venv_path).lower())


def test_install_packages_no_pip(tmpfolder, monkeypatch):
    tmp = Path(str(tmpfolder)).resolve()

    # Given pip is not installed
    monkeypatch.setattr(venv, "get_command", Mock(return_value=None))

    # when we run install_packages
    opts = {
        "project_path": tmp,
        "venv": tmp / ".venv",
        "venv_install": ["pytest>=6.0.0", "pip-tools"],
    }
    with pytest.raises(venv.NotInstalled) as ex:
        # Then it should throw an exception
        venv.install_packages({}, opts)

    assert "pip cannot be found" in str(ex)


@pytest.mark.slow
def test_api_with_venv(tmpfolder):
    venv_path = Path(tmpfolder) / "proj/.venv"
    # Given the venv does not exist yet
    assert not venv_path.exists()
    # when the CLI is invoked with --venv and --venv-install
    api.create_project(
        project_path="proj", extensions=[venv.Venv()], venv_install=["pytest>=6.0.0"]
    )
    # then the venv will be created accordingly
    assert venv_path.is_dir()
    # with python, pip and the installed executables
    assert list(venv_path.glob("*/python*"))
    assert list(venv_path.glob("*/pip*"))
    assert list(venv_path.glob("*/pytest*"))


@pytest.mark.slow
def test_cli_with_venv(tmpfolder):
    venv_path = Path(tmpfolder) / "proj/.venv"
    # Given the venv does not exist yet
    assert not venv_path.exists()
    # when the CLI is invoked with --venv and --venv-install
    cli.main(["proj", "--venv", "--venv-install", "pytest>=6.0.0"])
    # then the venv will be created accordingly
    assert venv_path.is_dir()
    # with python, pip and the installed executables
    assert list(venv_path.glob("*/python*"))
    assert list(venv_path.glob("*/pip*"))
    assert list(venv_path.glob("*/pytest*"))


@pytest.mark.slow
def test_cli_with_venv_and_pretend(tmpfolder):
    proj_path = Path(tmpfolder) / "proj"
    venv_path = proj_path / ".venv"
    # Given the venv does not exist yet
    assert not venv_path.exists()
    # when the CLI is invoked with --venv, --venv-install and --pretend
    cli.main(["proj", "--pretend", "--venv", "--venv-install", "pytest>=6.0.0"])
    # then the venv will not be created, or even the project itself
    assert not venv_path.exists()
    assert not proj_path.exists()


@pytest.mark.slow
def test_cli_with_venv_config(tmpfolder):
    venv_path = Path(tmpfolder) / "proj/.venv"
    # Given the venv does not exist yet
    assert not venv_path.exists()
    # and a config file exists containing venv
    config = """
    [snek]
    extensions =
        venv
    venv_install =
        pytest>=6.0.0
    """
    config_file = Path(tmpfolder, ".snek.cfg")
    config_file.write_text(cleandoc(config), encoding="utf-8")
    # when the CLI is invoked with that config file
    cli.main(["proj", "--config", str(config_file)])
    # then the venv will be created accordingly
    assert venv_path.is_dir()
    # with python, pip and the installed executables
    assert list(venv_path.glob("*/python*"))
    assert list(venv_path.glob("*/pip*"))
    assert list(venv_path.glob("*/pytest*"))



"""
Extension that removes configuration files for the Tox test automation tool.
"""

from typing import List

from .. import structure
from ..actions import Action, ActionParams, ScaffoldOpts, Structure
from . import Extension


class NoTox(Extension):
    """Prevent a tox configuration file from being created"""

    def activate(self, actions: List[Action]) -> List[Action]:
        """Activate extension, see :obj:`~snek.extension.Extension.activate`."""
        return self.register(actions, remove_files, after="define_structure")


def remove_files(struct: Structure, opts: ScaffoldOpts) -> ActionParams:
    """Remove .tox.ini file to structure"""
    return structure.reject(struct, "tox.ini"), opts



"""
Command-Line-Interface of snek
"""

import argparse
import logging
import sys
from pathlib import Path
from typing import List, Optional

from packaging.version import Version

from . import __version__ as snek_version
from . import api, templates
from .actions import ScaffoldOpts
from .actions import discover as discover_actions
from .exceptions import exceptions2exit
from .extensions import list_from_entry_points as list_all_extensions
from .identification import get_id
from .info import best_fit_license
from .log import ReportFormatter, logger
from .shell import shell_command_error2exit_decorator


def add_log_related_args(parser: argparse.ArgumentParser):
    """Add options that control verbosity/logger level"""
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_const",
        const=logging.INFO,
        dest="log_level",
        help="show additional information about current actions",
    )
    parser.add_argument(
        "-vv",
        "--very-verbose",
        action="store_const",
        const=logging.DEBUG,
        dest="log_level",
        help="show all available information about current actions",
    )
    parser.add_argument(
        "-P",
        "--pretend",
        dest="pretend",
        action="store_true",
        default=False,
        help="do not create project, but displays the log of all operations"
        " as if it had been created.",
    )


def add_init_args(parser: argparse.ArgumentParser):
    """Add the default options and arguments to the CLI parser."""

    # Here we can use api.DEFAULT_OPTIONS to provide the help text, but we should avoid
    # passing a `default` value to argparse, since that would shadow
    # `api.bootstrap_options`.
    # Setting defaults with `api.bootstrap_options` guarantees we do that in a
    # centralised manner, that works for both CLI and direct Python API invocation.

    parser.add_argument(
        "-n",
        "--name",
        dest="name",
        required=False,
        help="installable name "
        "(as in `pip install`/PyPI, default: basename of PROJECT_PATH)",
        metavar="NAME",
    )
    parser.add_argument(
        "-p",
        "--package",
        dest="package",
        required=False,
        help="package name (as in `import`, default: NAME)",
        metavar="PACKAGE_NAME",
    )
    parser.add_argument(
        "-d",
        "--description",
        dest="description",
        required=False,
        help="package description",
        metavar="TEXT",
    )
    license_choices = list(templates.licenses.keys())
    choices_help = ", ".join(license_choices)
    default_license = api.DEFAULT_OPTIONS["license"]
    parser.add_argument(
        "-l",
        "--license",
        dest="license",
        choices=license_choices,
        type=best_fit_license,
        required=False,
        help=f"package license like {choices_help} (default: {default_license})",
        metavar="LICENSE",
    )
    parser.add_argument(
        "-u",
        "--url",
        dest="url",
        required=False,
        help="main website/reference URL for package",
        metavar="URL",
    )
    parser.add_argument(
        "-f",
        "--force",
        dest="force",
        action="store_true",
        required=False,
        help="force overwriting an existing directory",
    )
    parser.add_argument(
        "-U",
        "--update",
        dest="update",
        action="store_true",
        required=False,
        help="update an existing project by replacing the most important files"
        " like setup.py etc. Use additionally --force to replace all scaffold files.",
    )

    # The following are basically for the CLI options, so having a default value is OK.
    parser.add_argument(
        "-V", "--version", action="version", version=f"snek {snek_version}"
    )
    add_log_related_args(parser)
    parser.add_argument(
        "--list-actions",
        dest="command",
        action="store_const",
        const=list_actions,
        help="do not create project, but show a list of planned actions",
    )


def add_extension_args(parser: argparse.ArgumentParser):
    """Add options and arguments defined by extensions to the CLI parser."""
    # load and instantiate extensions
    cli_extensions = list_all_extensions()

    for extension in cli_extensions:
        extension.augment_cli(parser)


def parse_args(args: List[str]) -> ScaffoldOpts:
    """Parse command line parameters respecting extensions

    Args:
        args: command line parameters as list of strings

    Returns:
        dict: command line parameters
    """
    # create the argument parser
    msg = "Snek is a tool for templating a Python project."
    parser = argparse.ArgumentParser(description=msg)
    parser.set_defaults(extensions=[], config_files=[])
    parsers = parser.add_subparsers(help='Sub commands')
    init_parser = parsers.add_parser('init', help="Initialize a new repository in the current directory")
    init_parser.set_defaults(func=init)
    add_init_args(init_parser)

    add_extension_args(parser)

    # Parse options and transform argparse Namespace object into common dict
    return _process_opts(vars(parser.parse_args(args)))


def _process_opts(opts: ScaffoldOpts) -> ScaffoldOpts:
    """Process and enrich command line arguments.

    Please not that there are many places where you can process scaffold options.
    This function should only be used when we absolutely need to be processed/corrected
    in the CLI-layer, before even touching the Python API (e.g. for configuring logging
    with the values given in the CLI).
    Default values should go to :obj:`snek.api.bootstrap_options` and derived
    values should go to :obj:`snek.actions.get_default_options`. This is important
    to keep feature parity between CLI and Python-only API.

    Args:
        opts: dictionary of parameters

    Returns:
        Dictionary of parameters from command line arguments
    """
    opts = {k: v for k, v in opts.items() if v not in (None, "")}
    # ^  Remove empty items, so we ensure setdefault works
    opts.setdefault("log_level", _default_log_level(opts))
    logger.reconfigure(opts)

    return opts


def get_log_level(args: Optional[List[str]] = None):
    """Get the configured log level directly by parsing CLI options from
    ``args`` or obj:`sys.argv`.

    Useful when the CLI crashes before applying the changes to the logger.
    """
    parser = argparse.ArgumentParser(add_help=False)
    add_log_related_args(parser)

    parsed, _remaining = parser.parse_known_args(args or sys.argv[:])
    partial_opts = vars(parsed)

    return partial_opts.get("log_level") or _default_log_level(partial_opts)


def _default_log_level(opts: ScaffoldOpts):
    # When pretending the user surely wants to see the output
    return logging.INFO if opts.get("pretend") else logging.WARNING


def init(opts: ScaffoldOpts):
    opts["project_name"] = Path('.').absolute().name
    run_scaffold(opts)


def run_scaffold(opts: ScaffoldOpts):
    """Actually scaffold the project, calling the python API

    Args:
        opts (dict): command line options as dictionary
    """
    api.create_project(opts)
    if opts["update"] and not opts["force"]:
        note = (
            "Update accomplished!\n"
            "Please check if your setup.cfg still complies with:\n"
            "https://snek.org/en/v{}/configuration.html"
        )
        base_version = Version(snek_version).base_version
        print(note.format(base_version))


def list_actions(opts: ScaffoldOpts):
    """Do not create a project, just list actions considering extensions

    Args:
        opts (dict): command line options as dictionary
    """
    actions = discover_actions(opts.get("extensions", []))

    print("Planned Actions:")
    for action in actions:
        print(ReportFormatter.SPACING + get_id(action))


def main(args: List[str]):
    """Main entry point for external applications

    Args:
        args: command line arguments
    """
    opts = parse_args(args)
    opts["func"](opts)


@shell_command_error2exit_decorator
@exceptions2exit([RuntimeError])
def run(args: Optional[List[str]] = None):
    """Entry point for console script"""
    main(args or sys.argv[1:])


if __name__ == "__main__":
    main(sys.argv[1:])
